import numpy as np
import torch as t
import pandas as pd
from joblib import Parallel, delayed
import matplotlib.pyplot as plt
from tqdm.autonotebook import tqdm
import logging
import random
from datetime import datetime, timedelta

from src.mmingest import MMSRIngest
from src.mmlookup import MMLookup
from src.mmutil import Util


class MMBet:
    def __init__(self, df_source):

        self.df_source = df_source[["Season", "DayNum", "Team Of Interest", "Opponent", "Mean", "Cover"]]
        self.df_source = df_source.rename(columns={"Mean": "pCover"})

    @classmethod
    @delayed
    def _gen_one_season(cls, df_source, season, season_length, gpd_dist):
        day_dfs = []
        for day in range(season_length):
            n_games = np.random.choice(gpd_dist)
            df_day = df_source.sample(n_games)
            df_day["DayNum"] = day
            day_dfs.append(df_day)
        df_season = pd.concat(day_dfs, ignore_index=True)
        df_season["Season"] = season

        return df_season

    @classmethod
    def generate_montecarlo_seasons(cls, df_source, n_seasons=1, season_length=130, games_per_day="random", seed=1234):
        np.random.seed(seed)

        if games_per_day == "random":
            gpd_dist = df_source.groupby(by=["Season", "DayNum"]).count().values[:, 0]
        else:
            gpd_dist = np.array(games_per_day)

        n_seasons = int(n_seasons)
        season_length = int(season_length)

        season_dfs = Parallel(n_jobs=-1)(
            cls._gen_one_season(df_source=df_source, season=season, season_length=season_length, gpd_dist=gpd_dist)
            for season in tqdm(range(n_seasons), desc=f"Generating {n_seasons} seasons")
        )

        return season_dfs

    @staticmethod
    def compound_bankroll_pct(df, future=False):
        e_bankroll = 1
        a_bankroll = 1
        df["Cumulative Expected BR %"] = e_bankroll
        df['Cumulative Actual BR %'] = 0 if future else a_bankroll

        pbar = tqdm(df.iterrows(), desc=f"ROI: {100 * (e_bankroll - 1): .02f}%", total=len(df))

        for idx, row in pbar:
            e_bankroll += e_bankroll * row["Expected BR % Increase"]
            a_bankroll += a_bankroll * row["Actual BR % Increase"]
            df.loc[idx, "Cumulative Expected BR %"] = e_bankroll
            df.loc[idx, "Cumulative Actual BR %"] = a_bankroll
            pbar.set_description(f"Expected ROI: {100 * (e_bankroll - 1): .02f}%", refresh=True)

        return df

    def kelly_bet(self, fraction=0.1, compound=False):
        df = self.df_source.copy()
        df["Wager (% of BR)"] = fraction * (df["pCover"] - (1. - df["pCover"]) / (1. / 1.1))
        df.loc[df["Wager (% of BR)"] <= 0, "Wager (% of BR)"] = 0
        df["%ROI"] = 0
        df.loc[df["Cover"] == 1, "%ROI"] = 2.1 / 1.1 - 1
        df.loc[df["Cover"] == 2, "%ROI"] = -1
        df["BR % Increase"] = df["%ROI"] * df["Wager (% of BR)"]
        df = df.groupby(by=["Season", "DayNum"], as_index=False).sum()[["Season", "DayNum", "Wager (% of BR)",
                                                                        "BR % Increase"]]

        if compound:
            df = self.compound_bankroll_pct(df)
        else:
            df["Cumulative BR %"] = df["BR % Increase"].cumsum()

        return df

    def efficient_frontier(self, method='tangent', risk_tolerance=None, compound=False, var_metric="semivariance", future=False):

        # confirm inputs are valid

        risk_tolerance = [0.05, 0.05, 0.05] if risk_tolerance is None else risk_tolerance
        assert var_metric in ["variance", "semivariance"]
        assert method in ['tangent', 'two_points', 'random']

        # get a data source with all the relevant global stats
        df_results = self.calc_global_stats(var_metric=var_metric)

        # create repository for results and length of the season loop
        season_returns_dfs = []
        season_details_dfs = []
        seasons = df_results["Season"].unique()

        # get the efficient frontier for every season in the data found in self.df_source
        for season in seasons:
            df_season = df_results[df_results["Season"] == season].copy()
            df_season = df_season.drop(columns="Season")
            days = df_season["DayNum"].unique()

            # create repository for daily returns and bet allocation details
            day_returns_dfs = []
            day_details_dfs = []

            # loop for each day
            for day in days:
                df_day = df_season[df_season["DayNum"] == day].copy()
                if method == 'tangent':
                    df_day_returns, df_day_details = self.tangent_ef(df_day=df_day, risk_tolerance=risk_tolerance, future=future)
                elif method == 'two_points':
                    df_day_returns, df_day_details = self.two_points_ef(df_day=df_day, risk_tolerance=risk_tolerance, future=future)
                elif method == 'random':
                    df_day_returns, df_day_details = self.random_ef(df_day=df_day, risk_tolerance=risk_tolerance, future=future)
                else:
                    logging.error(f'{method} is an unknown method!')
                day_returns_dfs.append(df_day_returns)
                day_details_dfs.append(df_day_details)

            df_season_returns = pd.concat(day_returns_dfs, ignore_index=True)
            df_season_returns["Season"] = season
            season_returns_dfs.append(df_season_returns)

            df_season_details = pd.concat(day_details_dfs, ignore_index=True)
            df_season_details["Season"] = season
            season_details_dfs.append(df_season_returns)

        df_returns = pd.concat(season_returns_dfs, ignore_index=True)
        if compound:
            df_returns = self.compound_bankroll_pct(df_returns)
        else:
            df_returns["Cumulative Expected BR %"] = df_returns["Expected BR % Increase"].cumsum()
            df_returns["Cumulative Actual BR %"] = df_returns["Actual BR % Increase"].cumsum()

        df_details = pd.concat(season_details_dfs, ignore_index=True)

        return df_returns, df_details

    def calc_global_stats(self, var_metric='semivariance'):
        df = self.df_source.copy()
        df["Actual ROR"] = 0
        df.loc[df["Cover"] == 1, "Actual ROR"] = 1. / 1.1
        df.loc[df["Cover"] == 2, "Actual ROR"] = -1
        df["Expected ROR"] = df["pCover"] * 1. / 1.1 - (1 - df["pCover"])

        if var_metric == "variance":
            df["Variance"] = df["pCover"] * (1 - df["pCover"])
        elif var_metric == "semivariance":
            df["Variance"] = ((1 - df["pCover"]) * 1.1) ** 2

        df = df[df["Expected ROR"] > 0]

        return df

    @classmethod
    def _prepare_results(cls, df_day=None, weights=None, exp_returns=None, variance=None, future=False):

        day = df_day.iloc[0]['DayNum']
        cash_weight = weights[0]
        exp_bankroll_increase = np.dot(weights, exp_returns)
        if future:
            df_day_returns = pd.DataFrame({"DayNum": day,
                                           "Wager (% of BR)": weights[1:].sum(),
                                           "Expected BR % Increase": exp_bankroll_increase}, index=[0])
        else:
            actual_returns = np.zeros_like(weights) if cash_weight == 1 else np.append(0, df_day["Actual ROR"].values)
            act_bankroll_increase = np.dot(weights, actual_returns)
            df_day_returns = pd.DataFrame({"DayNum": day,
                                           "Wager (% of BR)": weights[1:].sum(),
                                           "Actual BR % Increase": act_bankroll_increase,
                                           "Expected BR % Increase": exp_bankroll_increase}, index=[0])

        line = df_day.iloc[[0]].reset_index(drop=True)
        line['Team Of Interest'] = 'Cash Retained'
        for col in line.columns:
            if col not in ['Season', 'Segment', 'DayNum', 'Team Of Interest']:
                line[col] = '' if col == 'Opponent' else 0
                if col[:3] in ['T1_', 'T2_']:
                    line = line.drop(columns=[col], axis=1)
                    df_day = df_day.drop(columns=[col], axis=1)

        df_day_details = pd.concat([line, df_day]).reset_index(drop=True)
        df_day_details['Portfolio Weight'] = weights
        df_day_details['Weighted Expected ROR'] = df_day_details['Portfolio Weight'] * df_day_details['Expected ROR']
        df_day_details['Weighted Actual ROR'] = 0 if future else df_day_details['Portfolio Weight'] * df_day_details['Actual ROR']

        return df_day_returns, df_day_details

    @classmethod
    def tangent_ef(cls, df_day=None, risk_tolerance=0.05, future=False, allow_debt=False):

        n_bets = len(df_day)
        device = "cuda" if t.cuda.is_available() else "cpu"

        expected_returns = t.tensor(df_day["Expected ROR"].values, dtype=t.float32, device=device)
        variance_matrix = t.diag(t.tensor(df_day["Variance"].values, dtype=t.float32, device=device))
        inv_variance = t.inverse(variance_matrix)
        ones = t.ones((1, n_bets), dtype=t.float32, device=device)
        top = inv_variance @ expected_returns
        bottom = ones @ inv_variance @ expected_returns
        tangent_weights = top / bottom
        tangent_risk = t.sqrt(tangent_weights @ variance_matrix @ tangent_weights)

        if allow_debt:
            bet_comp = (risk_tolerance / tangent_risk).unsqueeze(0)
        else:
            bet_comp = (risk_tolerance / tangent_risk).unsqueeze(0) if tangent_risk >= risk_tolerance else t.tensor([1], dtype=t.float32, device=device)       # make sure we don't bet more than the bankroll

        p_optimal = t.cat([1 - bet_comp, tangent_weights * bet_comp])

        df_day_returns, df_day_details = cls._prepare_results(df_day=df_day, weights=p_optimal.cpu().detach().numpy(), exp_returns=np.append(0, expected_returns.cpu().detach().numpy()),
                                                              variance=np.append(0, variance_matrix.diagonal(0).cpu().detach().numpy()),
                                                              future=future)
        return df_day_returns, df_day_details

    @classmethod
    def tangent_ef_thresh(cls, df_day=None, bankroll=0, risk_tolerance=0.05, future=False, allow_debt=False, bet_thresh=1):

        # calculate portfolio weights and bet amounts
        betty = MMBet(df_source=df_day)
        df_day_all_bets = betty.calc_global_stats()
        returns, df_day_all_bets = betty.tangent_ef(df_day=df_day_all_bets, risk_tolerance=0.025, future=future, allow_debt=allow_debt)
        df_day_all_bets['Bet Amount'] = df_day_all_bets['Portfolio Weight'] * bankroll

        # get a df of today's bets that were allocated cash above the threshold
        cash_row = df_day_all_bets.loc[(df_day_all_bets['Team Of Interest'] == 'Cash Retained')]
        df_tan_no_cash = df_day_all_bets.loc[(df_day_all_bets['Bet Amount'] >= bet_thresh) & (df_day_all_bets['Team Of Interest'] != 'Cash Retained')]

        if len(df_tan_no_cash) != len(df_day_all_bets) - 1:
            df_day_merge = df_tan_no_cash.merge(df_day, on=['Season', 'Segment', 'DayNum', 'Team Of Interest', 'Opponent'], how='left', indicator=True)
            df_tan_no_cash = df_day_merge[cls._get_convert_cols(df_source=df_day_merge, df_target=df_day)].copy()
            df_tan_no_cash = cls._fix_col_names(df=df_tan_no_cash)
            returns, df_tan = cls.tangent_ef_thresh(df_day=df_tan_no_cash, bankroll=bankroll, risk_tolerance=risk_tolerance, future=future, allow_debt=allow_debt, bet_thresh=1)
        else:
            df_tan = pd.concat([cash_row, df_tan_no_cash])

        return returns, df_tan

    def two_points_ef(self, df_day=None, risk_tolerance=0.05, future=False, allow_debt=False):

        n_bets = len(df_day)
        v_rors = df_day['Expected ROR'].values
        v_weights = self._calc_weights(v_stds=np.sqrt(df_day['Variance'].values), v_rors=v_rors, n_bets=n_bets)

        if v_weights.sum() != 0:
            v_vars = np.sum(v_weights ** 2 * df_day['Variance'].values, axis=1)
            v_stds = np.sqrt(v_vars)
            v_w_sharpe = v_weights @ v_rors / v_stds
            optimal_idx = np.argmax(v_w_sharpe)
            sharpe_risk = v_stds[optimal_idx]
            sharpe_weights = v_weights[optimal_idx]

            if allow_debt:
                cash_bet_split = (risk_tolerance / sharpe_risk)
            else:
                cash_bet_split = (risk_tolerance / sharpe_risk) if sharpe_risk >= risk_tolerance else np.array([1], dtype=t.float32)

            p_optimal = np.empty(shape=(sharpe_weights.shape[0] + 1))
            p_optimal[0] = 1 - cash_bet_split
            p_optimal[1:] = sharpe_weights * cash_bet_split
            v_optimal_vars = v_weights ** 2 * df_day['Variance'].values
            v_optimal_vars = v_optimal_vars[optimal_idx]
        else:
            p_optimal = np.zeros(n_bets + 1)
            p_optimal[0] = 1
            v_optimal_vars = np.zeros(n_bets)

        df_day_returns, df_day_detail = self._prepare_results(df_day=df_day, weights=p_optimal, exp_returns=np.append(0, v_rors),
                                                              variance=np.append(0, v_optimal_vars), future=future)

        return df_day_returns, df_day_detail

    def random_ef(self, df_day=None, risk_tolerance=0.05, sample_size=1000, future=False):

        cash_bias = 1
        sample_size = int(sample_size)
        generator = np.random.default_rng(seed=42)
        device = "cuda" if t.cuda.is_available() else "cpu"
        alpha = np.append(cash_bias, df_day["pCover"].values)
        portfolios = t.tensor(generator.dirichlet(alpha, sample_size), dtype=t.float32, device=device)
        expected_returns = t.tensor(np.append(0, df_day["Expected ROR"].values), dtype=t.float32,
                                    device=device)
        return_vector = portfolios @ expected_returns
        variance_vector = t.tensor(df_day["Variance"].values, dtype=t.float32, device=device)
        risk_vector = t.sqrt((portfolios[:, 1:] ** 2) @ variance_vector)  # / (n_bets + 1)
        p_optimal = portfolios[t.argmax(return_vector[risk_vector < risk_tolerance])]

        df_day_returns, df_day_details = self._prepare_results(df_day=df_day,
                                                               weights=p_optimal.cpu().detach().numpy(),
                                                               exp_returns=(expected_returns * p_optimal).cpu().detach().numpy(),
                                                               variance=np.append(0, variance_vector.cpu().detach().numpy()),
                                                               future=future)
        return df_day_returns, df_day_details

    @classmethod
    def _get_convert_cols(cls, df_target=None, df_source=None):
        copy_cols = []
        source_cols = df_source.columns.tolist()
        target_cols = df_target.columns.tolist()
        for col in source_cols:
            if col in target_cols:
                copy_cols.append(col)
            elif col[:-2] in target_cols and col[-2:] == '_y':
                copy_cols.append(col)
            else:
                pass
                # print(f'Skipping col: {col} > {col[:-2]} > {col[-2:]}')

        return copy_cols

    @classmethod
    def _fix_col_names(cls, df=None):
        for col in df.columns.tolist():
            if col[-2:] == '_y':
                df[col[:-2]] = df[col]
                df = df.drop(col, axis=1)
            elif col == 'pCover':
                df['Mean'] = df[col]
                df = df.drop(col, axis=1)
            else:
                # print(f'Skipping {col}...')
                pass
        return df

    @classmethod
    def _calc_weights(cls, v_stds=None, v_rors=None, n_bets=None):

        v_ones = np.ones(shape=n_bets)
        weights = []

        a = v_ones @ v_rors
        b = np.transpose(v_rors) @ v_rors
        c = np.transpose(v_ones) @ v_ones
        d = b * c - a ** 2
        g = (1 / (d + 1e-9)) * (b * v_ones - a * v_rors)
        h = (1 / (d + 1e-9)) * (c * v_rors - a * v_ones)

        w_max = g + h * v_rors.max()
        w_min = g + h * v_rors.min()

        sd_max = int(np.sqrt(v_stds.max()) + 1)

        alphas = np.linspace(start=-sd_max, stop=sd_max, num=2*sd_max*100)

        for alpha in alphas:
            weights.append(alpha * w_max + (1 - alpha) * w_min)

        return np.array(weights)

    @classmethod
    def plot_ef(cls, x=None, y=None, alpha=0.5, x_dim='Volatility', y_dim='Expected_Return', output_path=None):

        fig, ax = plt.subplots(figsize=(10, 5))
        ax.scatter(x, y, alpha=0.70)
        ax.set_title('Efficient Frontier')
        ax.set_xlabel(x_dim)
        ax.set_ylabel(y_dim)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.grid(color='blue', linestyle='-', linewidth=0.25, alpha=alpha)
        # plt.show()
        fig.savefig(output_path, dpi=320)

    @classmethod
    def get_random_day(cls, df_source=None):
        seasons = df_source['Season'].unique()
        season = random.choice(seasons)
        df_season = df_source.loc[df_source['Season'] == season]
        days = df_season['DayNum'].unique()
        day = random.choice(days)
        df_day = df_season.loc[df_season['DayNum'] == day]
        # df_day = df_source.loc[(df_source['Season'] == 2005) & (df_source['DayNum'] == 74)]

        return df_day

    @classmethod
    def get_bet_record(cls, df_bet_sheet, helper_path="./data/our_data/"):
        season_lookup = MMLookup(source='seasons')

        if "T1_TeamID" not in df_bet_sheet.columns:
            teams_lookup = MMLookup(source='team_spellings')
            df_bet_sheet["T1_TeamID"] = df_bet_sheet.apply(
                lambda x: int(teams_lookup.key_to_val("TeamID", x["Team Of Interest"].lower())), axis=1)
            df_bet_sheet["T2_TeamID"] = df_bet_sheet.apply(
                lambda x: int(teams_lookup.key_to_val("TeamID", x["Opponent"].lower())), axis=1)

        season_days = df_bet_sheet.apply(lambda x: f"{x['Season']}_{x['DayNum']}", axis=1).unique()
        day_result_dfs = []

        for sd in season_days:
            season, day_num = sd.split("_")
            day_zero = season_lookup.key_to_val("DayZero", int(season))
            game_date = (datetime.strptime(day_zero, "%m/%d/%Y").date() + timedelta(days=int(day_num))).isoformat()

            day_ingestor = MMSRIngest(game_date, helper_path=helper_path)
            day_result_dfs.append(Util.swap_extend(day_ingestor.compact_results_df))

        df_results = pd.concat(day_result_dfs, ignore_index=True)
        df_bet_record = df_bet_sheet.merge(df_results[["T1_TeamID", "T1_Score", "T2_TeamID", "T2_Score"]], how="left",
                                           on=["T1_TeamID", "T2_TeamID"])

        df_bet_record = df_bet_record[["Season", "Segment", "DayNum", "Team Of Interest", "T1_Score", "Opponent",
                                       "T2_Score", "aVSpread", "pCover", "Portfolio Weight", "Bet Amount"]]

        df_bet_record = df_bet_record.rename(columns={"T1_Score": "TOI Score", "T2_Score": "Opponent Score"})

        df_bet_record["Actual ROR"] = 0
        df_bet_record.loc[df_bet_record["TOI Score"] - df_bet_record["Opponent Score"] > df_bet_record["aVSpread"],
                          "Actual ROR"] = 1/1.1
        df_bet_record.loc[df_bet_record["TOI Score"] - df_bet_record["Opponent Score"] < df_bet_record["aVSpread"],
                          "Actual ROR"] = -1.
        df_bet_record["BR % Change"] = df_bet_record["Actual ROR"] * df_bet_record["Portfolio Weight"]
        df_bet_record["Winnings"] = df_bet_record["Actual ROR"] * df_bet_record["Bet Amount"]

        return df_bet_record
