# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import logging
import os

import h2o
import pandas as pd
from joblib import Parallel, delayed, wrap_non_picklable_objects
from tqdm.autonotebook import tqdm

import src.mmfeature as mf
import src.mmutil as mu

# ---------------------------------------------------------------------------------------------------------
#  These columns are those which are used to exclude them from the train set with config.exclude=True
#  D_COLS is for "detail" files
#  C_COLS is for "compact" files
#  You can also set config.exclude=False, and then pass a column list that is include only
# ---------------------------------------------------------------------------------------------------------
D_COLS = ['Season', 'Segment', 'T1_TeamID', 'T2_TeamID', 'location', 'T1_Score', 'T2_Score', 'DayNum', 'NumOT',
          'T1_FGM', 'T1_FGA', 'T1_FGM3', 'T1_FGA3', 'T1_FTM', 'T1_FTA', 'T1_OR', 'T1_DR', 'T1_Ast', 'T1_TO', 'T1_Stl',
          'T1_Blk', 'T1_PF', 'T2_FGM', 'T2_FGA', 'T2_FGM3', 'T2_FGA3', 'T2_FTM', 'T2_FTA', 'T2_OR', 'T2_DR', 'T2_Ast',
          'T2_TO', 'T2_Stl', 'T2_Blk', 'T2_PF']

C_COLS = ['Season', 'Segment', 'T1_TeamID', 'T2_TeamID', 'location', 'T1_Score', 'T2_Score', 'DayNum', 'NumOT']


# -----------------------------------------------------------------------------
#
#   Name:      MMData
#
#   Purpose:   Primary data object. Opens, prepares, and annotates
#              train, test, and validation data sets.
#
#
#   Usage:     data = MMData(data_path=path_to_csv, start_yr=2003, end_yr=2020)
#              df_train = data.get_train_data()
#
# -----------------------------------------------------------------------------
class MMData:
    def __init__(self, config):

        self.config = config

        # Open the data frame to build the train/test/predict sets from
        self.df_core = mu.Util.get_source_data(self.config.source, self.config.data_path, local_root=self.config.local_root, gcs_root=self.config.gcs_root)
        self.df_core = self.df_core.sort_values(by=["Season", "DayNum"], ignore_index=True)
        self.df_core = self.df_core.reset_index(drop=True)

        # chunk seasons and df['Segment'] field to df_core
        self.df_core = mu.Util.chunk_seasons(self.df_core, self.config.segment_count)

        # here we split train/test (typically both drawn from df_core). however, march_mad is an exception
        if self.config.test_source != "march_mad":
            self.df_core_train, self.df_core_test = mu.Util.split(self.df_core,
                                                                  start=config.train_start,
                                                                  end=config.train_end,
                                                                  start_cut=config.test_start,
                                                                  end_cut=config.test_end,
                                                                  segment_count=config.segment_count)
        else:
            self.df_core_train, _ = mu.Util.split(self.df_core,
                                                  start=config.train_start,
                                                  end=config.train_end,
                                                  start_cut=config.test_start,
                                                  end_cut=config.test_end,
                                                  segment_count=config.segment_count)

            self.df_core_test = mu.Util.get_source_data(self.config.test_source, path=self.config.data_path, local_root=self.config.local_root, gcs_root=self.config.gcs_root)
            self.df_core_test["Segment"] = self.config.segment_count

        # df_pred is assumed to be cropped and segmented
        if self.config.future:
            self.df_core_pred = self.config.df_pred
            self.df_core_pred['Segment'] = mu.Util.adjust_date(config.train_end, segment_count=config.segment_count)[1]

    # --------------------------------------------------------------
    #  Gets a training data set
    # --------------------------------------------------------------
    def get_train_data(self):

        # get the file name to use
        base, ext = os.path.splitext(self.config.csv_name)
        csv_name = base + '_Train' + ext

        # get the data
        df = self.get_data(df_source=self.df_core_train, target_mode='train', csv_name=csv_name, future=False)

        return df

    # --------------------------------------------------------------
    #  Gets a test data set (uses the same inputs as training)
    # --------------------------------------------------------------
    def get_test_data(self):

        # get the file name to use
        base, ext = os.path.splitext(self.config.csv_name)
        csv_name = base + '_Test' + ext

        # get the data
        df = self.get_data(df_source=self.df_core_test, target_mode='test', csv_name=csv_name, future=False)

        return df

    # --------------------------------------------------------------
    #  Gets a kaggle submission set for prediction
    # --------------------------------------------------------------
    def get_pred_data(self):

        # get the file name to use
        base, ext = os.path.splitext(self.config.csv_name)
        csv_name = base + '_Pred' + ext

        # get the data
        df = self.get_data(df_source=self.df_core_pred, csv_name=csv_name, future=True)

        return df

    # --------------------------------------------------------------
    #  Core data prep pipeline
    # --------------------------------------------------------------
    def get_data(self, target_mode=None, df_source=None, csv_name='', future=False):

        # make the date tuples
        start = (df_source.iloc[0].Season, df_source.iloc[0].Segment, df_source.iloc[0].DayNum)
        end = (df_source.iloc[-1].Season, df_source.iloc[-1].Segment, df_source.iloc[-1].DayNum)

        logging.info(f'Building a dataset for the period {start[0]} | S{start[1]} | D{start[2]}' +
                     f' >> {end[0]} | S{end[1]} | D{end[2]}')

        if self.config.refresh:
            logging.info('Swapping and extending...')
            df = mu.Util.swap_extend(df_source)

            # build df
            df = self._preprocess_features(df, start, end)

            # get targets (the predicted y)
            logging.info('Creating targets...')
            df = self._generate_targets(df, target_mode=target_mode, future=future)

            # if desired, save a copy for reuse
            if self.config.to_csv:
                logging.info(f'Saving {self.config.local_root}{self.config.data_path}{csv_name}')
                df.to_csv(f'{self.config.local_root}{self.config.data_path}{csv_name}', index=False)
        else:
            # otherwise, load a saved version of the train data
            df = pd.read_csv(f'{self.config.local_root}{self.config.data_path}{csv_name}', encoding='ISO-8859-1')
            logging.info(f'Last version of {csv_name} loaded...')

        # support output of h2o frames
        df = self.preprocessor_to_h2o(df=df)

        return df

    # --------------------------------------------------------------
    #  Runs the feature_def/metric pipeline (parallelized for speed)
    # --------------------------------------------------------------
    def _preprocess_features(self, df, start=None, end=None):

        feature_def_list = mf.MMFeatureDefinition().build_feature_list(self.config.feature_list, start, end)

        n_seasons = end[0] - start[0] + 1

        pbar = tqdm(total=n_seasons*len(self.config.feature_list),
                    desc="Generating Features",
                    bar_format="{l_bar}{bar}|", position=0, leave=True)

        features = Parallel(n_jobs=self.config.njobs, prefer="threads")(
            self._generate_feature(feature, start, end, pbar) for feature, start, end in feature_def_list)

        # Add resulting features to df
        for feature in features:
            df = feature.add_feature(df)

        return df

    # --------------------------------------------------------------
    #  Preprocesses the final pipeline and converts to h2o frame
    # --------------------------------------------------------------
    def preprocessor_to_h2o(self, df):

        if self.config.exclude:
            drop_columns = [c for c in self.config.train_cols if c in df.columns]
            save_columns = drop_columns
            drop_columns = drop_columns + [self.config.target_col]
        else:
            drop_columns = [self.config.target_col]
            save_columns = self.config.train_cols

        target = df[self.config.target_col]
        df_pairings = df[save_columns]  # save to add back after preprocessor (needed for analysis)
        df_features = df.drop(columns=drop_columns)

        # Run features through preprocessor pipeline and then reattach column names
        features = self.config.preprocessor.fit_transform(df_features)
        df_features = pd.DataFrame(data=features,
                                   columns=self.config.preprocessor.get_feature_names_out(df_features.columns))

        # Reattach targets and pairings to df_features then convert to H2OFrame
        df = df_features.assign(**{self.config.target_col: target})
        df = df.join(df_pairings)

        if self.config.to_h2o:
            data = h2o.H2OFrame(df)

            return data
        else:
            return df

    # --------------------------------------------------------------
    #  Splits the keys from Kaggle style submission file csv
    # --------------------------------------------------------------
    @classmethod
    def split_keys(cls, row):
        keys = pd.Series([row.ID.split('_')[0], row.ID.split('_')[1], row.ID.split('_')[2], row.ID.split('_')[3]])
        return keys

    # --------------------------------------------------------------
    #  Calculates the target column using any MMTeamTransform
    # --------------------------------------------------------------
    def _generate_targets(self, df, target_mode=None, future=False):

        if future:
            df[self.config.target_col] = 1
        else:
            if target_mode == 'train':
                df_source = mu.Util.get_source_data(self.config.train_target_source, path=self.config.data_path)
                if self.config.train_target_source == "spreads":
                    df = self._get_spread_targets(df, df_source)
                else:
                    df[self.config.target_col] = self.config.target_op(df)
            elif target_mode == 'test':
                df_source = mu.Util.get_source_data(self.config.test_target_source, path=self.config.data_path)
                if self.config.test_target_source == "spreads":
                    df = self._get_spread_targets(df, df_source)
                else:
                    df[self.config.target_col] = self.config.target_op(df)
            else:
                logging.error(f"Unrecognized target mode: {target_mode}. Must be either 'train' or 'test' unless "
                              f"future=True (in which case target_mode is ignored)...")

            df = df.dropna(subset=[self.config.target_col])
            df = df.reset_index(drop=True)

        return df

    # --------------------------------------------------------------
    #  Runs the MMMetric -> MMFeature pipeline
    # --------------------------------------------------------------
    @delayed
    @wrap_non_picklable_objects
    def _generate_feature(self, feature_def, start, end, pbar):

        logging.debug(
            f"Calculating team feature: {feature_def['feature_params']['feature_name']} >>> metric: "
            f"{feature_def['feature_params']['metric_name'] if 'metric_name' in feature_def['feature_params'].keys() else None}...")

        # hackathon save source for the feature_def so that you can reset it to the orig after you build the feature_def
        orig_src = feature_def['feature_params']['source']


        # Add the feature_def to the data
        feature = feature_def['feature_obj'](**feature_def['feature_params'],
                                             metric_params = feature_def['feature_params'],
                                             start=start,
                                             end=end,
                                             local_root = self.config.local_root,
                                             gcs_root=self.config.gcs_root,
                                             segment_count=self.config.segment_count,
                                             save_path=self.config.feature_path,
                                             data_path=self.config.data_path,
                                             pbar=pbar)

        # hackathon save source for the feature_def, resetting it to the orig after building the feature_def
        feature_def['feature_params']['source'] = orig_src

        return feature

    def _get_spread_targets(self, df, df_source):

        df_source = mu.Util.swap_extend_spreads(df_source)
        df = df.merge(df_source[["Season", "DayNum", "T1_TeamID", "T2_TeamID", "line"]],
                      on=["Season", "DayNum", "T1_TeamID", "T2_TeamID"], how="left")
        df[self.config.target_col] = self.config.target_op(df)
        df = df.drop(columns=["line"])

        return df

