# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import csv
import logging
import os.path

import numpy as np
import pandas as pd
import statsmodels.api as sm

import src.mmmetric as mm
import src.mmutil as mu
from src.mmlookup import MMLookup


# ----------------------------------------------------------------------------------------------------------------------
#
#   Name:        MMFeatureDefinition
#
#   Purpose:     A minimal class meant to serve as a repository of features from which to draw
#
#   Lags and Windows: These only apply to MMFeatureGLM features.
#
#                     window = the number of seasons to include in the calculation of this feature_def/metric
#
#                     lag = the number of seasons prior to end the train/test/predict target season to start
#                           the calc of this feature_def
#
#   Examples:
#
#           lag = 0, window = 0: In a typical train on 2018 regular season, predict 2018 march madness where
#                                a feature_def is calculated solely for the 2018 regular season.  Or train on
#                                2021-2022 games played to date, predict 2022 games to come.
#
#           lag = 0, window = 4: Train on 2018 regular season, predict 2018 march madness where
#                                a feature_def is calculated as a single feature_def for 2015 - 2018
#
#           lag = 0, window = 4: Train on 2018 regular season, predict 2018 march madness where
#                                a feature_def is calculated as a single feature_def for 2015 - 2018
#
# ----------------------------------------------------------------------------------------------------------------------
class MMFeatureDefinition:
    # ------------------------------------------------------------------------
    # gets a new feature_def
    # ------------------------------------------------------------------------
    @classmethod
    def get_feature_def(cls, feature, source=None, lag=None, window=None):

        # These features require a lag and window, set to inputs or default to L1, W4
        lag_win_features = ['Team_Q', 'Off_Q', 'Def_Q', 'Tempo', 'Off_Poss_Q', 'Def_Poss_Q', 'Ast_Ratio', 'TO_Ratio',
                            'True_Shooting', 'Eff_FG', 'OR_Pct', 'DR_Pct', 'TR_Pct']
        if feature in lag_win_features:
            if lag is None:
                lag = 1
            if window is None:
                window = 4

        # long if/elif defines the dict for each known feature_def
        if feature == 'Seed':
            if source is None:
                source = 'seeds'
            feature_def = {
                'feature_obj': MMFeature,

                'feature_params': {
                    'feature_name': 'Seed',
                    'metric_name': 'Seed',
                    'source': source,
                    'metric_obj': mm.MMMetricSeed
                }
            }

        elif feature == 'aVSpread':
            if source is None:
                source = 'spreads'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'aVSpread',
                    'metric_name': 'Spread',
                    'source': source,
                    'metric_obj': mm.MMMetricSpread,
                    'op': mm.MMTeamTransform.spread
                }
            }

        elif feature == 'pVSpread':
            if source is None:
                source = 'pVSpread'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'pVSpread',
                    'metric_name': 'pVSpread',
                    'source': source,
                    'metric_obj': mm.MMMetricPred
                }
            }

        elif feature == 'pMOV_b5':
            if source is None:
                source = 'pMOV_b5'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'pMOV_b5',
                    'metric_name': 'pMOV_b5',
                    'source': source,
                    'metric_obj': mm.MMMetricPred
                }
            }

        elif feature == 'pVSpread_abs':
            if source is None:
                source = 'pVSpread_abs'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'pVSpread_abs',
                    'metric_name': 'pVSpread_abs',
                    'source': source,
                    'metric_obj': mm.MMMetricPred
                }
            }

        elif feature == 'pMOV_b5_abs':
            if source is None:
                source = 'pMOV_b5_abs'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'pMOV_b5_abs',
                    'metric_name': 'pMOV_b5_abs',
                    'source': source,
                    'metric_obj': mm.MMMetricPred
                }
            }
        elif feature == 'pMOV_b3':
            if source is None:
                source = 'pMOV_b3'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'pMOV_b3',
                    'metric_name': 'pMOV_b3',
                    'source': source,
                    'metric_obj': mm.MMMetricPred
                }
            }

        elif feature == 'pCover_B5Abs':
            if source is None:
                source = 'pCover_B5Abs'
            feature_def = {
                'feature_obj': MMFeatureKeyVal,

                'feature_params': {
                    'feature_name': 'pCover_B5Abs',
                    'metric_name': 'pCover_B5Abs',
                    'source': source,
                    'metric_obj': mm.MMMetricPred
                }
            }

        elif feature == 'Location':
            if source is None:
                source = 'cur_season'
            feature_def = {
                'feature_obj': MMFeatureCategorical,

                'feature_params': {
                    'feature_name': 'Location',
                    'metric_name': 'Location',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': 0,
                    'window': 0,
                    'op': mm.MMTeamTransform.location
                }
            }

        elif feature == 'Team_Q':
            if source is None:
                source = 'cur_season'
            feature_def = {
                'feature_obj': MMFeatureGLM,

                'feature_params': {
                    'feature_name': 'Team_Q',
                    'metric_name': 'MOV',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': lag,
                    'window': window,
                    'op': mm.MMTeamTransform.mov,
                    'family': 'Gaussian',
                    'formula': 'Team_Q~-1+T1_TeamID-T2_TeamID'
                }
            }

        elif feature == 'Off_Q':
            if source is None:
                source = 'cur_season'
            feature_def = {
                'feature_obj': MMFeatureGLM,

                'feature_params': {
                    'feature_name': 'Off_Q',
                    'metric_name': 'Off_Score',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': lag,
                    'window': window,
                    'op': mm.MMTeamTransform.off_score,
                    'family': 'Gaussian',
                    'formula': 'Off_Q~-1+T1_TeamID+T2_TeamID'
                }
            }

        elif feature == 'Def_Q':
            if source is None:
                source = 'cur_season'
            feature_def = {
                'feature_obj': MMFeatureGLM,

                'feature_params': {
                    'feature_name': 'Def_Q',
                    'metric_name': 'Def_Score',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': lag,
                    'window': window,
                    'op': mm.MMTeamTransform.def_score,
                    'family': 'Gaussian',
                    'formula': 'Def_Q~-1+T1_TeamID+T2_TeamID'
                }
            }

        elif feature == 'Tempo':
            if source is None:
                source = 'cur_detail'
            feature_def = {
                'feature_obj': MMFeatureGLM,

                'feature_params': {
                    'feature_name': 'Tempo',
                    'metric_name': 'Poss',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': lag,
                    'window': window,
                    'op': mm.MMTeamTransform.adj_possessions,
                    'family': 'Gaussian',
                    'formula': 'Tempo~-1+T1_TeamID+T2_TeamID'
                }
            }

        elif feature == 'Off_Poss_Q':
            if source is None:
                source = 'cur_detail'
            feature_def = {
                'feature_obj': MMFeatureGLM,

                'feature_params': {
                    'feature_name': 'Off_Poss_Q',
                    'metric_name': 'Off_Rtg',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': lag,
                    'window': window,
                    'op': mm.MMTeamTransform.off_rtg,
                    'family': 'Gaussian',
                    'formula': 'Off_Poss_Q~-1+T1_TeamID+T2_TeamID'
                }
            }

        elif feature == 'Def_Poss_Q':
            if source is None:
                source = 'cur_detail'
            feature_def = {
                'feature_obj': MMFeatureGLM,

                'feature_params': {
                    'feature_name': 'Def_Poss_Q',
                    'metric_name': 'Def_Rtg',
                    'source': source,
                    'metric_obj': mm.MMMetricGame,
                    'lag': lag,
                    'window': window,
                    'op': mm.MMTeamTransform.def_rtg,
                    'family': 'Gaussian',
                    'formula': 'Def_Poss_Q~-1+T1_TeamID+T2_TeamID'
                }
            }

        elif feature == 'M_Rank':
            if source is None:
                source = 'ordinals'
            feature_def = {
                'feature_obj': MMFeatureOrdinal,

                'feature_params': {
                    'feature_name': 'M_Rank',
                    'source': source
                }
            }

        elif feature == "Ast_Ratio":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "Ast_Ratio",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "Ast_Ratio",
                    "op": mm.MMTeamTransform.assist_ratio,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }

        elif feature == "TO_Ratio":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "TO_Ratio",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "TO_Ratio",
                    "op": mm.MMTeamTransform.turnover_ratio,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }
        elif feature == "True_Shooting":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "True_Shooting",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "True_Shooting",
                    "op": mm.MMTeamTransform.true_shooting,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }

        elif feature == "Eff_FG":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "Eff_FG",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "Eff_FG",
                    "op": mm.MMTeamTransform.effective_fg_pct,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }
        elif feature == "OR_Pct":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "OR_Pct",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "OR_Pct",
                    "op": mm.MMTeamTransform.off_reb_pct,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }

        elif feature == "DR_Pct":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "DR_Pct",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "DR_Pct",
                    "op": mm.MMTeamTransform.def_reb_pct,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }

        elif feature == "TR_Pct":
            if source is None:
                source = "cur_detail"
            feature_def = {
                "feature_obj": MMFeature,
                "feature_params": {
                    "feature_name": "TR_Pct",
                    "source": source,
                    "metric_obj": mm.MMMetricSeason,
                    "metric_name": "TR_Pct",
                    "op": mm.MMTeamTransform.tot_reb_pct,
                    "aggregation": "sum",
                    "lag": lag,
                    "window": window
                }
            }

        elif feature == "Abs_Pts":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_Pts",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_Pts",
                    "op": mm.MMPlayerTransform.points,
                    "window": window
                }
            }

        elif feature == "Abs_OR":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_OR",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_OR",
                    "op": mm.MMPlayerTransform.off_reb,
                    "window": window
                }
            }

        elif feature == "Abs_DR":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_DR",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_DR",
                    "op": mm.MMPlayerTransform.def_reb,
                    "window": window
                }
            }

        elif feature == "Abs_Ast":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_Ast",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_Ast",
                    "op": mm.MMPlayerTransform.assists,
                    "window": window
                }
            }

        elif feature == "Abs_TO":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_TO",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_TO",
                    "op": mm.MMPlayerTransform.turnovers,
                    "window": window
                }
            }

        elif feature == "Abs_Blk":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_Blk",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_Blk",
                    "op": mm.MMPlayerTransform.blocks,
                    "window": window
                }
            }

        elif feature == "Abs_Stl":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_Stl",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_Stl",
                    "op": mm.MMPlayerTransform.steals,
                    "window": window
                }
            }

        elif feature == "Abs_PF":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_PF",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_PF",
                    "op": mm.MMPlayerTransform.fouls,
                    "window": window
                }
            }

        elif feature == "Abs_MP":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_MP",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_MP",
                    "op": mm.MMPlayerTransform.minutes,
                    "window": window
                }
            }

        elif feature == "Abs_EFG":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_EFG",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_EFG",
                    "op": mm.MMPlayerTransform.effective_fg_pct,
                    "window": window
                }
            }

        elif feature == "Abs_TS":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_TS",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_TS",
                    "op": mm.MMPlayerTransform.true_shooting,
                    "window": window
                }
            }

        elif feature == "Abs_PIE":
            if source is None:
                source = "absences"
            feature_def = {
                "feature_obj": MMFeatureAbsence,
                "feature_params": {
                    "feature_name": "Abs_PIE",
                    "source": source,
                    "metric_obj": mm.MMMetricAbsence,
                    "metric_name": "Abs_PIE",
                    "op": mm.MMPlayerTransform.pie,
                    "window": window
                }
            }

        else:
            feature_def = -1
            logging.error(f"'{feature}' is not a known feature!")

        return feature_def

    # ------------------------------------------------------------------------
    # Used by MMData to build a feature_def pipeline for preprocess_features()
    # ------------------------------------------------------------------------
    @classmethod
    def build_feature_list(cls, feature_list, start, end):

        # lags and windows are relative to the start and end of the dataset being built
        # start and end are each a tuple (Season, Segment, DayNum)

        starts = []
        ends = []
        features = []

        for feature in feature_list:
            features.append(feature)
            starts.append(start)
            ends.append(end)

        feature_set = zip(features, starts, ends)

        return feature_set


# -----------------------------------------------------------------------------
#
#   Name:      MMFeature
#
#   Purpose:   Basic feature_def addition object.  Used to calculate stats from
#              a df, and then merge the result with a train or test set.
#
#
#   Usage:     feature = MMFeature()
#              df_train = feature_def.add_feature(path=path_to_csv, source='rs')
#
# -----------------------------------------------------------------------------
class MMFeature:
    def __init__(self, *args, **kwargs):

        # these are required attributes for all features
        self.feature_name = kwargs.pop('feature_name')
        self.source = kwargs.pop('source')
        self.start = kwargs.pop('start')
        self.end = kwargs.pop('end')
        self.segment_count = kwargs.pop('segment_count')
        self.family = kwargs.pop('family') if 'family' in kwargs.keys() else None
        self.formula = kwargs.pop('formula') if 'formula' in kwargs.keys() else None
        self.lag = kwargs.pop('lag') if 'lag' in kwargs.keys() else None
        self.window = kwargs.pop('window') if 'window' in kwargs.keys() else None
        self.metric_params = kwargs.pop('metric_params')
        self.gcs_root = kwargs.pop('gcs_root') if 'gcs_root' in kwargs.keys() else None
        self.local_root = kwargs.pop('local_root') if 'local_root' in kwargs.keys() else None
        self.save_path = kwargs.pop('save_path')
        self.data_path = kwargs.pop('data_path')

        self.save_path = os.path.join(self.local_root, self.save_path)

        self.pbar = kwargs.pop('pbar')

        self.total_segments = ((self.end[0] - self.start[0]) * self.segment_count) + (self.end[1] - self.start[1])

        self._set_df_source()

        # load the feature
        self.df_feature = self._load_feature()

    def _set_df_source(self):
        if isinstance(self.source, pd.DataFrame):
            # here we assume if we're passed a df its already cropped and segmented
            self.df_source = self.source
            self.covid_skip = False
        else:
            # This is the source data for your feature_def
            self.covid_skip = True if self.source == 'march_mad' else False
            self.df_source = mu.Util.get_source_data(self.source, path=self.data_path, local_root=self.local_root, gcs_root=self.gcs_root)
            self.df_source = self.df_source.sort_values(by=["Season", "DayNum"], ignore_index=True)
            self.df_source = self.df_source.reset_index(drop=True)
            self.df_source = mu.Util.chunk_seasons(self.df_source, segment_count=self.segment_count)
            self.df_source = mu.Util.crop(self.df_source, self.start, self.end)

    # ------------------------------------------------------------------------
    # Adds a feature to the passed df
    # ------------------------------------------------------------------------
    def add_feature(self, df):
        # you must return a df_train to the caller with the originally
        # supplied df + your feature_def added to the end for both T1 and T2
        df = self._merge_feature(df)

        return df

    # ------------------------------------------------------------------------
    # Empty method that can be replaced by subclasses (see comments below)
    # ------------------------------------------------------------------------
    def _calc_team_feature(self):
        # use this method to calculate a feature_def for addition to the primary data set. Typically, the df passed out
        # of this function is a df ready for joining to the train set,
        pass

    # ------------------------------------------------------------------------
    # Empty method that can be replaced by subclasses (see comments below)
    # ------------------------------------------------------------------------
    def _add_team_feature(self, df):
        # use this method to do the attaching of a calculated feature_def
        # to the primary data set
        return df

    # ------------------------------------------------------------------------
    # Merges the feature with the df
    # ------------------------------------------------------------------------
    def _merge_feature(self, df):
        # This code takes care of adding the feature_def you calculated above to the main data set

        if df["Segment"].max() >= self.segment_count:
            df_feature_mm = self.df_feature[self.df_feature["Segment"] == 0].copy()
            df_feature_mm["Segment"] = self.segment_count
            df_feature_mm["Season"] = df_feature_mm["Season"] - 1

            self.df_feature = self.df_feature.append(df_feature_mm)

        self.df_feature = self.df_feature.drop_duplicates(subset=['Season', 'Segment', 'TeamID'])

        df = pd.merge(df, self.df_feature, how='left', left_on=['Season', 'Segment', 'T1_TeamID'],
                      right_on=['Season', 'Segment', 'TeamID']).drop('TeamID', axis=1)
        df[self.feature_name] = df[self.feature_name].fillna(df[self.feature_name].median())
        df = df.rename(columns={self.feature_name: 'T1_' + self.feature_name})

        # denormalize and add feature_def info to the main df for T2
        df = pd.merge(df, self.df_feature, how='left', left_on=['Season', 'Segment', 'T2_TeamID'],
                      right_on=['Season', 'Segment', 'TeamID']).drop('TeamID', axis=1)
        df[self.feature_name] = df[self.feature_name].fillna(df[self.feature_name].median())
        df = df.rename(columns={self.feature_name: 'T2_' + self.feature_name})

        return df

    # ------------------------------------------------------------------------
    # Saves the feature to disk
    # ------------------------------------------------------------------------
    def _save_feature(self, df, season, segment):

        relative_path = f"data/{self.feature_name}_{season}_{segment}-{self.segment_count}" \
                        f"_L{self.lag}_W{self.window}.csv"
        save_path = os.path.join(self.save_path, relative_path)
        df.loc[:, ["TeamID", self.feature_name]].to_csv(save_path, index=False)

        new_file_data = [
            self.feature_name,
            self.segment_count,
            season,
            segment,
            self.lag,
            self.window,
            relative_path
        ]

        return new_file_data

    def _load_feature_db(self, path):
        dtypes = {
            "FeatureName": str,
            "SegmentCount": "Int64",
            "Season": "Int64",
            "Segment": "Int64",
            "Lag": "Int64",
            "Window": "Int64",
            "FilePath": str
        }

        if os.path.exists(path):
            df = pd.read_csv(path, dtype=dtypes)

        else:
            df = pd.DataFrame(columns=["FeatureName", "SegmentCount", "Season", "Segment", "Lag", "Window", "FilePath"])
            df.to_csv(path, index=False)

            os.makedirs(os.path.join(self.save_path, "data"))

        if self.lag is None or self.window is None:
            df = df.loc[(df["FeatureName"] == self.feature_name) &
                        (df["SegmentCount"] == self.segment_count), :]

        else:
            df = df.loc[(df["FeatureName"] == self.feature_name) &
                        (df["SegmentCount"] == self.segment_count) &
                        (df["Lag"] == self.lag) &
                        (df["Window"] == self.window), :]

        return df

    def _get_season_seg_count(self, season):

        if season == self.end[0]:
            segments = range(self.end[1] + 1)

        elif season == self.start[0]:
            segments = range(self.start[1], self.segment_count)

        else:
            segments = range(self.segment_count + 1)

        return segments

    # ------------------------------------------------------------------------
    # Loads the feature from disk
    # ------------------------------------------------------------------------
    def _load_feature(self):

        # get the feature database
        db_path = os.path.join(self.save_path, "feature_database.csv")
        df_ref_db = self._load_feature_db(db_path)

        # create a list store all segment feature_dfs
        segment_dfs = []

        # open the feature_database and update it as necessary
        with open(db_path, "a", newline="") as db:
            db_writer = csv.writer(db)

            # for each season we look for an existing feature file or keep on
            for season in range(self.start[0], self.end[0] + 1):

                if not self.covid_skip:     # << this covid skip logic exists solely for march_mad training/prediction

                    # get a list of this season's feature files
                    logging.debug(f'Building {self.feature_name} for {season}...')
                    df_season_db = df_ref_db.loc[df_ref_db["Season"] == season, :]

                    # get the number of segments in this season
                    segments = self._get_season_seg_count(season)

                    for segment in segments:

                        # if there is no entry in feature_database for this Season | Segment, we have to build one
                        if df_season_db.loc[df_season_db["Segment"] == segment, :].empty:

                            # if a metric is needed as df_source we build it otherwise, we copy in the self.df_source
                            if self.metric_params['metric_name'] is not None:
                                metric = self.metric_params['metric_obj'](**self.metric_params, start=(season, segment),
                                                                          end=(season, segment),
                                                                          segment_count=self.segment_count,
                                                                          path=self.data_path,
                                                                          gcs_root=self.gcs_root,
                                                                          local_root= self.local_root)
                                df_source = metric.get_metric()
                            else:
                                df_source = self.df_source.copy()

                            # add the season and segment to the source data
                            df_source = df_source.loc[(df_source['Season'] == season) &
                                                      (df_source["Segment"] == segment)]

                            # as long as df_source has data, we can add the feature and update the feature database
                            if not df_source.empty:
                                self.df_source = df_source
                                df = self._add_team_feature(df_source)
                                db_writer.writerow(self._save_feature(df, season, segment))

                                # append this feature for this Season | Segment pair to the segment_df list
                                segment_dfs.append(df)

                            # otherwise, there is no data to build the feature (march_mad 2020
                            else:
                                logging.debug(f'Season | Segment: {season} | {segment} has not and cannot be made...')
                        else:

                            # here we pick up on features that were already built and in the database
                            logging.debug(f'{self.feature_name} was previously calculated for {season}, segment '
                                          f'{segment}/{self.segment_count} << using that...')

                            relative_path = df_season_db.loc[df_season_db["Segment"] == segment, "FilePath"].values[0]
                            df = pd.read_csv(os.path.join(self.save_path, relative_path))
                            df["Season"] = season
                            df["Segment"] = segment

                            # append this feature for this Season | Segment pair to the segment_df list
                            segment_dfs.append(df)

                if self.pbar is not None:
                    self.pbar.update(1)

        df_feature = pd.concat(segment_dfs, ignore_index=True)

        return df_feature


# -----------------------------------------------------------------------------
#
#   Name:      MMFeatureKeyVal
#
#               this class is meant to handle features that are ??unswapped?? and stored in
#               'Season, DayNum, T1, T2, Feature' form
# -----------------------------------------------------------------------------

class MMFeatureKeyVal(MMFeature):
    def _set_df_source(self):
        if isinstance(self.source, pd.DataFrame):
            self.df_source = self.source
            self.covid_skip = False
        else:
            # This is the source data for your feature_def
            self.covid_skip = True if self.source == 'march_mad' else False
            self.df_source = mu.Util.get_source_data(self.source, path=self.data_path, local_root=self.local_root, gcs_root=self.gcs_root)

    def _merge_feature(self, df):
        df = df.merge(self.df_feature[["Season", "DayNum", "T1_TeamID", "T2_TeamID", self.feature_name]],
                      on=["Season", "DayNum", "T1_TeamID", "T2_TeamID"], how="left")

        df = df.drop_duplicates(subset=["Season", "DayNum", "T1_TeamID", "T2_TeamID"])

        df[self.feature_name].fillna(df[self.feature_name].median(), inplace=True)
        df = df.dropna()

        return df

    # ------------------------------------------------------------------------
    # Saves the feature to disk
    # ------------------------------------------------------------------------
    def _save_feature(self, df, season, segment):

        relative_path = f"data/{self.feature_name}_{season}_{segment}-{self.segment_count}.csv"
        save_path = os.path.join(self.save_path, f"data/{self.feature_name}_{season}_"
                                                 f"{segment}-{self.segment_count}.csv")
        df.loc[:, ["DayNum", "T1_TeamID", "T2_TeamID", self.feature_name]].to_csv(save_path, index=False)

        new_file_data = [
            self.feature_name,
            self.segment_count,
            season,
            segment,
            "",
            "",
            relative_path
        ]

        return new_file_data


class MMFeatureCategorical(MMFeatureKeyVal):

    def _merge_feature(self, df):
        df = df.merge(self.df_feature[["Season", "DayNum", "T1_TeamID", "T2_TeamID", self.feature_name]],
                      on=["Season", "DayNum", "T1_TeamID", "T2_TeamID"], how="left")

        df = df.drop_duplicates(subset=["Season", "DayNum", "T1_TeamID", "T2_TeamID"])

        return df


class MMFeatureAbsence(MMFeatureKeyVal):
    def _merge_feature(self, df):
        if self.df_feature.empty:
            df[f"T1_{self.feature_name}"] = 0
            df[f"T2_{self.feature_name}"] = 0
        else:
            df = pd.merge(df, self.df_feature, how="left", left_on=["Season", "DayNum", "T1_TeamID"],
                          right_on=["Season", "DayNum", "TeamID"]).drop(columns="TeamID")
            df[self.feature_name] = df[self.feature_name].fillna(0)
            df = df.rename(columns={
                self.feature_name: f"T1_{self.feature_name}"
            })

            df = pd.merge(df, self.df_feature, how="left", left_on=["Season", "DayNum", "T2_TeamID"],
                          right_on=["Season", "DayNum", "TeamID"]).drop(columns="TeamID")
            df[self.feature_name] = df[self.feature_name].fillna(0)
            df = df.rename(columns={
                self.feature_name: f"T2_{self.feature_name}"
            })

        return df

    # ------------------------------------------------------------------------
    # Saves the feature_def to disk
    # ------------------------------------------------------------------------
    def _save_feature(self, df, season):
        relative_path = f"data/{self.feature_name}_{season}_W{self.window}.csv"
        save_path = os.path.join(self.save_path, relative_path)
        df.loc[:, ["DayNum", "TeamID", self.feature_name]].to_csv(save_path, index=False)

        new_file_data = [
            self.feature_name,
            "",
            season,
            "",
            "",
            self.window,
            relative_path
        ]

        return new_file_data

    def _load_feature(self):

        db_path = os.path.join(self.save_path, "feature_database.csv")

        df_ref_db = self._load_feature_db(db_path)

        if self.window is None:
            df_ref_db = df_ref_db.loc[(df_ref_db["FeatureName"] == self.feature_name), :]
        else:
            df_ref_db = df_ref_db.loc[(df_ref_db["FeatureName"] == self.feature_name) &
                                      (df_ref_db["Window"]) == self.window, :]

        season_dfs = []

        with open(db_path, "a", newline="") as db:
            db_writer = csv.writer(db)

            for season in range(self.start[0], self.end[0] + 1):
                df_season_db = df_ref_db.loc[df_ref_db["Season"] == season, :]

                if df_season_db.empty:
                    logging.debug(f'Building {self.feature_name} for {season}...')

                    # df_source = self.df_source.copy()
                    # df_source = df_source.loc[(df_source['Season'] == season)]

                    if self.metric_params['metric_name'] is not None:
                        metric = self.metric_params['metric_obj'](**self.metric_params,
                                                                  start=(season, 0),
                                                                  end=(season, self.segment_count),
                                                                  segment_count=self.segment_count,
                                                                  path=self.data_path)
                        df_source = metric.get_metric()
                    else:
                        df_source = self.df_source.copy()

                    df_source = df_source.loc[(df_source['Season'] == season)] if not df_source.empty else df_source

                    if not df_source.empty:
                        df = self._add_team_feature(df_source)
                        db_writer.writerow(self._save_feature(df, season))
                        season_dfs.append(df)

                    else:
                        logging.debug(f'Season | Segment: {season} has not and cannot be made...')
                else:
                    logging.debug(f'{self.feature_name} was previously calculated for {season} << using that...')
                    relative_path = df_season_db["FilePath"].values[0]
                    df = pd.read_csv(os.path.join(self.save_path, relative_path))
                    df["Season"] = season
                    season_dfs.append(df)

                self.pbar.update(1)

        if len(season_dfs) > 0:
            df_feature = pd.concat(season_dfs, ignore_index=True)
        else:
            df_feature = pd.DataFrame()

        return df_feature


# -----------------------------------------------------------------------------
#
#   Name:      MMFeatureOrdinal
#
#   Purpose:   Creates an MMFeature out of the Massey ordinals and adds it
#              and then merges the result with a train or test set.
#
#
#   Usage:     feature_def = MMFeatureOrdinal()
#              df_train = feature_def.add_feature(data_path=path_to_csv, source='rs')
#
# -----------------------------------------------------------------------------
class MMFeatureOrdinal(MMFeatureKeyVal):

    def __init__(self, *args, **kwargs):
        self.systems = ['COL', 'DOL', 'MOR', 'POM', 'SAG', 'WLK']
        self.cols = [f"T1_{sys}" for sys in self.systems] + [f"T2_{sys}" for sys in self.systems]
        super().__init__(self, *args, **kwargs)

    def _add_team_feature(self, df):
        season_lookup = MMLookup(source='seasons')

        df = df[df["SystemName"].isin(self.systems)]
        df = df.drop_duplicates(subset=["Season", "DayNum", "TeamID", "SystemName"])
        df = df.pivot(index=["Season", "DayNum", "TeamID"], columns="SystemName", values="OrdinalRank").reset_index()

        season_days = df.apply(lambda x: f"{int(x.Season)}{int(x.DayNum)}", axis=1)
        season_days = season_days.unique()

        for i, sd in enumerate(season_days[:-1]):
            season = int(sd[:4])
            day_num = int(sd[4:])

            next_season = int(season_days[i + 1][:4])
            next_day = int(season_days[i + 1][4:])

            if next_season > season:
                next_day = season_lookup.key_to_val("MaxDayNum", next_season) + 1

            season_teams = df.loc[df["Season"] == season, "TeamID"].unique()
            n_teams = len(season_teams)

            df_filler = pd.DataFrame({
                "Season": season,
                "DayNum": np.repeat(np.arange(day_num + 1, next_day), n_teams),
                "TeamID": np.tile(season_teams, next_day - day_num - 1)
            }, dtype=int)

            df_last_rank = pd.DataFrame(columns=["TeamID"])
            df_last_rank["TeamID"] = season_teams

            for sys in self.systems:
                last_rank = df.loc[((df[sys].notna()) &
                                    (df["DayNum"] == df.loc[df["DayNum"] <= day_num, "DayNum"].max())), ["TeamID", sys]]

                if not last_rank.empty:
                    df_last_rank = df_last_rank.merge(right=last_rank, on="TeamID", how="left")
                else:
                    df_last_rank[sys] = np.nan

            df_filler = df_filler.merge(right=df_last_rank, on="TeamID", how="left").sort_values(by="DayNum",
                                                                                                 ignore_index=True)

            df = pd.concat([df[((df["DayNum"] <= day_num) & (df["Season"] == season)) | (df["Season"] < season)],
                            df_filler,
                            df[((df["DayNum"] >= next_day) & (df["Season"] == season)) | (df["Season"] > season)]])

        return df

    # ------------------------------------------------------------------------
    # Saves the feature_def to disk
    # ------------------------------------------------------------------------
    def _save_feature(self, df, season):
        relative_path = f"data/{self.feature_name}_{season}.csv"
        save_path = os.path.join(self.save_path, relative_path)
        df.loc[:, ["DayNum", "TeamID"] + self.systems].to_csv(save_path, index=False)

        new_file_data = [
            self.feature_name,
            "",
            season,
            "",
            "",
            "",
            relative_path
        ]

        return new_file_data

    # ------------------------------------------------------------------------
    # Loads the feature_def from disk
    # ------------------------------------------------------------------------
    def _load_feature(self):

        db_path = os.path.join(self.save_path, "feature_database.csv")

        df_ref_db = self._load_feature_db(db_path)

        df_ref_db = df_ref_db.loc[(df_ref_db["FeatureName"] == self.feature_name), :]

        season_dfs = []

        with open(db_path, "a", newline="") as db:
            db_writer = csv.writer(db)

            for season in range(self.start[0], self.end[0] + 1):
                df_season_db = df_ref_db.loc[df_ref_db["Season"] == season, :]

                if df_season_db.empty:
                    logging.debug(f'Building {self.feature_name} for {season}...')
                    df_source = self.df_source.copy()
                    df_source = df_source.loc[(df_source['Season'] == season)]
                    df = self._add_team_feature(df_source)
                    db_writer.writerow(self._save_feature(df, season))
                else:
                    logging.debug(f'{self.feature_name} was previously calculated for {season} << using that...')
                    relative_path = df_season_db["FilePath"].values[0]
                    df = pd.read_csv(os.path.join(self.save_path, relative_path))
                    df["Season"] = season

                season_dfs.append(df)
                self.pbar.update(1)

        df_feature = pd.concat(season_dfs, ignore_index=True)

        return df_feature


# -----------------------------------------------------------------------------
#
#   Name:      MMFeatureGLM
#
#   Purpose:   General Linear Model feature_def addition object.  Used to calculate stats from
#              a df, and then merge the result with a train or test set.
#              Currently only supports the regular season file and those lke it.
#
#
#   Usage:     feature_def = MMFeatureGLM()
#              TODO: Document parameters
#              df_train = feature_def.add_feature(data_path=path_to_csv, source='rs')
#
# -----------------------------------------------------------------------------
class MMFeatureGLM(MMFeature):
    def _set_df_source(self):
        if isinstance(self.source, pd.DataFrame):
            # here we assume if we're passed a df its already cropped and segmented
            self.df_source = self.source
            self.covid_skip = False

        else:
            # This is the source data for your feature_def
            self.covid_skip = True if self.source == 'march_mad' else False
            self.df_source = mu.Util.get_source_data(self.source, path=self.data_path, gcs_root=self.gcs_root, local_root=self.local_root)
            self.df_source = self.df_source.sort_values(by=["Season", "DayNum"], ignore_index=True)
            self.df_source = self.df_source.reset_index(drop=True)
            self.df_source = mu.Util.chunk_seasons(self.df_source, segment_count=self.segment_count)

    @classmethod
    def _add_teamid_dummy_feature(cls, df):

        df['T1_TeamID'] = df['T1_TeamID'].astype(str)
        df['T2_TeamID'] = df['T2_TeamID'].astype(str)

        return df

    def _calc_team_feature(self, df, start, end):

        data = mu.Util.crop(df, start=start, end=end)
        logging.debug(f'GLM Feature: {self.feature_name} using data from: {start} to: {end} returned {len(data)}')
        if self.family == 'Binomial':
            glm = sm.GLM.from_formula(formula=self.formula, data=data, family=sm.families.Binomial()).fit()

        elif self.family == 'Gaussian':
            glm = sm.GLM.from_formula(formula=self.formula, data=data, family=sm.families.Gaussian()).fit()

        else:
            logging.error('Invalid GLM family!')

        # extracting parameters from glm
        df_feature = pd.DataFrame(glm.params).reset_index()
        df_feature.columns = ['TeamID', 'Beta']

        feature_end = mu.Util.adjust_date(end, adj=self.lag, segment_count=self.segment_count)

        df_feature['Season'] = feature_end[0]
        df_feature['Segment'] = feature_end[1]

        # taking exp due to binomial model being used
        if self.family == 'Binomial':
            df_feature[self.feature_name] = np.exp(df_feature['Beta'])
        else:
            df_feature = df_feature.rename(columns={'Beta': self.feature_name})

        # only interested in glm parameters with T1_, as T2_ should be mirroring T1_ ones
        df_feature = df_feature.loc[df_feature.TeamID.str.contains('T1_')].reset_index(drop=True)
        df_feature['TeamID'] = df_feature['TeamID'].apply(lambda x: x[10:14]).astype(int)

        # TODO: do we still need this
        df_feature = df_feature.drop_duplicates(subset=["TeamID", "Season", "Segment"])

        return df_feature

    def _add_team_feature(self, df, window_start, window_end):

        df = self._add_teamid_dummy_feature(df)

        # calculate window-based quality metric
        df = self._calc_team_feature(df, window_start, window_end)

        return df

    # ------------------------------------------------------------------------
    # Loads the feature_def from disk
    # ------------------------------------------------------------------------
    def _load_feature(self):

        db_path = os.path.join(self.save_path, "feature_database.csv")

        df_ref_db = self._load_feature_db(db_path)

        df_ref_db = df_ref_db.loc[(df_ref_db["FeatureName"] == self.feature_name) &
                                  (df_ref_db["SegmentCount"] == self.segment_count) &
                                  (df_ref_db["Lag"] == self.lag) &
                                  (df_ref_db["Window"] == self.window), :]

        segment_dfs = []

        with open(db_path, "a", newline="") as db:
            db_writer = csv.writer(db)

            for season in range(self.start[0], self.end[0] + 1):
                if not self.covid_skip:
                    logging.debug(f'Building {self.feature_name} for {season}...')
                    df_season_db = df_ref_db.loc[df_ref_db["Season"] == season, :]

                    if season == self.end[0]:
                        segments = range(self.end[1] + 1)
                    elif season == self.start[0]:
                        segments = range(self.start[1], self.segment_count)
                    else:
                        segments = range(self.segment_count)

                    for segment in segments:
                        if df_season_db.loc[df_season_db["Segment"] == segment, :].empty:

                            logging.debug(f'Season: {season} | Segment: {segment} not found in feature database...')
                            window_start, window_end = mu.Util.\
                                adjust_for_lag_and_window(start=(season, segment),
                                                          end=(season, segment),
                                                          lag=self.lag,
                                                          window=self.window,
                                                          segment_count=self.segment_count)

                            if self.metric_params['metric_name'] is not None:
                                metric = self.metric_params['metric_obj'](**self.metric_params,
                                                                          start=window_start,
                                                                          end=window_end,
                                                                          segment_count=self.segment_count,
                                                                          path=self.data_path,
                                                                          local_root=self.local_root,
                                                                          gcs_root=self.gcs_root)
                                df_source = metric.get_metric()
                            else:
                                logging.error(f"GLM feature {self.feature_name} passed with metric object set to "
                                              f"{self.metric_params['metric_obj']}!!!")

                            if not df_source.empty:
                                df = self._add_team_feature(df_source, window_start, window_end)
                                db_writer.writerow(self._save_feature(df, season, segment))
                                segment_dfs.append(df)
                            else:
                                logging.debug(f'Season | Segment: {season} | {segment} has not and cannot be made...')
                        else:
                            logging.debug(f'{self.feature_name} was previously calculated for {season}, segment '
                                          f'{segment}/{self.segment_count} << using that...')

                            relative_path = os.path.join(self.save_path,
                                                         df_season_db.loc[df_season_db["Segment"] == segment,
                                                                          'FilePath'].values[0])

                            df = pd.read_csv(relative_path)
                            df["Season"] = season
                            df["Segment"] = segment
                            segment_dfs.append(df)

                    self.pbar.update(1)

        df_feature = pd.concat(segment_dfs, ignore_index=True)

        return df_feature
