# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import json
import logging
from datetime import datetime, date, timedelta
from urllib.request import Request, urlopen
import os
import time
import pandas as pd
import requests
from joblib import Parallel, delayed
from lxml.etree import ParserError
from sportsipy.ncaab.boxscore import Boxscores, Boxscore
from sportsipy.ncaab.teams import Teams
from tqdm.autonotebook import tqdm

from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions
from webdriver_manager.chrome import ChromeDriverManager

from src.mmlookup import MMLookup
from src.mmutil import Util


# --------------------------------------------------------------------------------------------------------------
#
#   Name:      MMSRIngest
#
#   Purpose:   Object for retrieving data over a given game_date range from
#              sportsreference.com and putting it in the format used by Kaggle.
#
#
#   Usage:     ingestor = MMSRIngest(start_date="2021-11-09", end_date="2021-12-04", helper_path=path_to_helper_csvs)
#              df_compact = data.compact_results_df
#              df_detailed = data.detailed_results_df
#
# --------------------------------------------------------------------------------------------------------------
class MMSRIngest:
    def __init__(self, start_date, end_date=None, helper_path="../data/our_data/", march_madness=False, n_jobs=-2,
                 local_root='./', gcs_root='ops/'):

        self.start_date = date.fromisoformat(start_date)
        self.local_root = local_root
        self.gcs_root = gcs_root

        if end_date is not None:
            self.end_date = date.fromisoformat(end_date)
            logging.info(f"Fetching boxscores from {self.start_date} to {self.end_date}...")
            self.boxscores = Boxscores(date=self.start_date, end_date=self.end_date)

        else:
            logging.info(f"Fetching boxscores from {self.start_date}")
            self.boxscores = Boxscores(date=self.start_date)
            self.end_date = self.start_date

        self.march_madness = march_madness

        self._team_spell = MMLookup(source='team_spellings')
        self._seasons = MMLookup(source='seasons')
        self._helper_path = helper_path
        self.n_jobs = n_jobs

    # Scrapes results between self.start_date and self.end_date and returns a dataframe
    # in the same format as *CompactResults.csv
    @property
    def compact_results_df(self):
        day_dfs = []

        for day, games in self.boxscores.games.items():
            if not games:
                continue

            df_day = pd.DataFrame(games)
            df_day = df_day.loc[~df_day["non_di"]]

            if self.march_madness:
                df_day = df_day.loc[df_day["top_25"]]

            # Sportsreference.com classifies stats by home and away, so to convert to Kaggle format games won by the
            # home team and away team need to be handled differently
            home_winners = df_day.loc[df_day["home_abbr"] == df_day["winning_abbr"]]
            home_winners = home_winners[["home_abbr", "home_score", "away_abbr", "away_score"]]
            home_winners = home_winners.rename(columns={
                "home_abbr": "WTeamID",
                "home_score": "WScore",
                "away_abbr": "LTeamID",
                "away_score": "LScore"
            })
            home_winners["WLoc"] = "H"

            away_winners = df_day.loc[df_day["away_abbr"] == df_day["winning_abbr"]]
            away_winners = away_winners[["away_abbr", "away_score", "home_abbr", "home_score"]]
            away_winners = away_winners.rename(columns={
                "away_abbr": "WTeamID",
                "away_score": "WScore",
                "home_abbr": "LTeamID",
                "home_score": "LScore"
            })
            away_winners["WLoc"] = "A"

            game_date = datetime.strptime(day, "%m-%d-%Y").date()

            year, day_num = self.convert_game_date(game_date)

            df_day_kaggle = pd.concat([home_winners, away_winners])
            df_day_kaggle = df_day_kaggle.assign(Season=year, DayNum=day_num, NumOT=0)  # NumOT is not practical to get

            day_dfs.append(df_day_kaggle)

        df = pd.concat(day_dfs, ignore_index=True)
        df = df[["Season", "DayNum", "WTeamID", "WScore", "LTeamID", "LScore", "WLoc", "NumOT"]]
        df["WTeamID"] = df.apply(lambda x: self._team_spell.key_to_val("TeamID", x["WTeamID"]), axis=1)
        df["LTeamID"] = df.apply(lambda x: self._team_spell.key_to_val("TeamID", x["LTeamID"]), axis=1)

        if self.march_madness:
            df["WLoc"] = "N"

        return df

    # Internal method to process games from a given day returned by Boxscores and convert into Kaggle
    # format. Used by detailed_results_df for parallel processing.
    @delayed
    def _process_day_results(self, day, games):
        game_dfs = []

        for game in games:
            if game["non_di"]:
                continue

            if self.march_madness and not game["top_25"]:
                continue

            bs_obj = Boxscore(game["boxscore"])

            try:
                bs = bs_obj.dataframe
            except TypeError:
                continue

            if bs is None:
                continue

            if bs["winner"].values[0] == "Home":
                winner = "home"
                loser = "away"

            else:
                winner = "away"
                loser = "home"

            bs = bs.rename(columns={
                f"{winner}_points": "WScore",
                f"{loser}_points": "LScore",
                f"winner": "WLoc",
                f"{winner}_field_goals": "WFGM",
                f"{winner}_field_goal_attempts": "WFGA",
                f"{winner}_three_point_field_goals": "WFGM3",
                f"{winner}_three_point_field_goal_attempts": "WFGA3",
                f"{winner}_free_throws": "WFTM",
                f"{winner}_free_throw_attempts": "WFTA",
                f"{winner}_offensive_rebounds": "WOR",
                f"{winner}_defensive_rebounds": "WDR",
                f"{winner}_assists": "WAst",
                f"{winner}_turnovers": "WTO",
                f"{winner}_steals": "WStl",
                f"{winner}_blocks": "WBlk",
                f"{winner}_personal_fouls": "WPF",
                f"{loser}_field_goals": "LFGM",
                f"{loser}_field_goal_attempts": "LFGA",
                f"{loser}_three_point_field_goals": "LFGM3",
                f"{loser}_three_point_field_goal_attempts": "LFGA3",
                f"{loser}_free_throws": "LFTM",
                f"{loser}_free_throw_attempts": "LFTA",
                f"{loser}_offensive_rebounds": "LOR",
                f"{loser}_defensive_rebounds": "LDR",
                f"{loser}_assists": "LAst",
                f"{loser}_turnovers": "LTO",
                f"{loser}_steals": "LStl",
                f"{loser}_blocks": "LBlk",
                f"{loser}_personal_fouls": "LPF"
            })

            bs["WTeamID"] = game[f"{winner}_abbr"]
            bs["LTeamID"] = game[f"{loser}_abbr"]

            game_dfs.append(bs)

        game_date = datetime.strptime(day, "%m-%d-%Y").date()
        year, day_num = self.convert_game_date(game_date)

        df_day = self._aggregate_game_dfs(game_dfs, year, day_num)

        return df_day

    # Scrapes results between self.start_date and self.end_date and returns a dataframe in the same
    # format as *DetailedResults.csv
    @property
    def detailed_results_df(self):
        day_dfs = Parallel(n_jobs=self.n_jobs)(
            self._process_day_results(day, games) for day, games in tqdm(self.boxscores.games.items(),
                                                                         desc=f"Processing games from {self.start_date}"
                                                                              f" to {self.end_date}"))

        df = pd.concat(day_dfs, ignore_index=True)
        df = df.dropna(how='all')  # drop all non-D1 games
        df["WTeamID"] = df.apply(lambda x: int(self._team_spell.key_to_val("TeamID", x["WTeamID"].lower())), axis=1)
        df["LTeamID"] = df.apply(lambda x: int(self._team_spell.key_to_val("TeamID", x["LTeamID"].lower())), axis=1)

        df.loc[df["WLoc"] == "Away", "WLoc"] = "A"
        df.loc[df["WLoc"] == "Home", "WLoc"] = "H"
        df.loc[df["WLoc"] == "Neutral", "WLoc"] = "N"

        df["NumOT"] = 0  # NumOT is not practical to get

        if self.march_madness:
            df["WLoc"] = "N"

        df = df[['Season', 'DayNum', 'WTeamID', 'WScore', 'LTeamID', 'LScore', 'WLoc', 'NumOT', 'WFGM', 'WFGA', 'WFGM3',
                 'WFGA3', 'WFTM', 'WFTA', 'WOR', 'WDR', 'WAst', 'WTO', 'WStl', 'WBlk', 'WPF', 'LFGM', 'LFGA', 'LFGM3',
                 'LFGA3', 'LFTM', 'LFTA', 'LOR', 'LDR', 'LAst', 'LTO', 'LStl', 'LBlk', 'LPF']]

        return df

    # Returns a dataframe of games between start_date and end_date without any results or features, just Season, DayNum,
    # WTeamID, and LTeamID
    @property
    def prediction_df(self):
        days = []

        for day, games in self.boxscores.games.items():
            df_day = pd.DataFrame(games)
            df_day = df_day.loc[~df_day["non_di"]]

            day = datetime.strptime(day, "%m-%d-%Y").date()

            year, day_num = self.convert_game_date(day)

            df_day["away_abbr"] = df_day.apply(
                lambda x: int(self._team_spell.key_to_val("TeamID", x["away_abbr"].lower())), axis=1)
            df_day["home_abbr"] = df_day.apply(
                lambda x: int(self._team_spell.key_to_val("TeamID", x["home_abbr"].lower())), axis=1)

            df_day["DayNum"] = day_num
            df_day["Season"] = year

            df_day = df_day.rename(columns={
                "away_abbr": "WTeamID",
                "home_abbr": "LTeamID"
            })

            df_day["WLoc"] = "A"

            days.append(df_day)

        df = pd.concat(days)

        df = df[["Season", "DayNum", "WTeamID", "LTeamID", "WLoc"]]

        return df

    @delayed
    def _process_day_players(self, day, games):
        game_dfs = []

        for game in games:

            if game["non_di"]:
                continue

            if self.march_madness and not game["top_25"]:
                continue

            if game["boxscore"] == '':
                continue

            try:
                bs = Boxscore(game["boxscore"])
            except (IndexError, ParserError):
                continue

            home_dfs = [p.dataframe for p in bs.home_players]
            away_dfs = [p.dataframe for p in bs.away_players]
            player_dfs = home_dfs + away_dfs

            if len(player_dfs) > 0:
                df_game = pd.concat(player_dfs)
            else:
                continue

            df_game = df_game.reset_index()

            df_game = df_game.rename(columns={
                "index": "PlayerID",
                "assists": "Ast",
                "blocks": "Blk",
                "defensive_rebounds": "DR",
                "field_goal_attempts": "FGA",
                "field_goals": "FGM",
                "free_throw_attempts": "FTA",
                "free_throws": "FTM",
                "minutes_played": "MP",
                "offensive_rebounds": "OR",
                "personal_fouls": "PF",
                "points": "Pt",
                "steals": "Stl",
                "three_point_attempts": "FGA3",
                "three_pointers": "FGM3",
                "turnovers": "TO",
            })

            df_game = df_game

            game_dfs.append(df_game)

        game_date = datetime.strptime(day, "%m-%d-%Y").date()
        year, day_num = self.convert_game_date(game_date)

        df_day = self._aggregate_game_dfs(game_dfs, year, day_num)

        return df_day

    # Scrapes player boxscores between self.start_date and self.end_date and returns a dataframe
    @property
    def player_stats_df(self):
        day_dfs = Parallel(n_jobs=self.n_jobs)(
            self._process_day_players(day, games) for day, games in tqdm(self.boxscores.games.items(),
                                                                         desc=f"Processing games from {self.start_date}"
                                                                              f" to {self.end_date}"))

        df = pd.concat(day_dfs, ignore_index=True)

        df = df[['Season', 'DayNum', 'PlayerID', 'Pt', 'FGM', 'FGA', 'FGM3', 'FGA3', 'FTM', 'FTA', 'OR', 'DR', 'Ast',
                 'TO', 'Stl', 'Blk', 'PF', "MP"]]

        return df

    @delayed
    def _process_day_absences(self, day, games, rosters):
        game_dfs = []

        game_date = datetime.strptime(day, "%m-%d-%Y").date()
        year, day_num = self.convert_game_date(game_date)

        for game in games:

            if game["non_di"]:
                continue

            if self.march_madness and not game["top_25"]:
                continue

            if game["boxscore"] == '':
                continue

            try:
                bs = Boxscore(game["boxscore"])
            except (IndexError, ParserError):
                continue

            home_id = self._team_spell.key_to_val("TeamID", game["home_abbr"].lower())
            away_id = self._team_spell.key_to_val("TeamID", game["away_abbr"].lower())

            home_roster = rosters[str(year)][str(home_id)]
            away_roster = rosters[str(year)][str(away_id)]

            home_players = [p.player_id for p in bs.home_players]
            away_players = [p.player_id for p in bs.away_players]

            df_home = pd.DataFrame({
                "TeamID": home_id,
                "Absences": list(set(home_roster).difference(home_players))
            })

            df_away = pd.DataFrame({
                "TeamID": away_id,
                "Absences": list(set(away_roster).difference(away_players))
            })

            game_dfs.extend([df_home, df_away])

        df_day = self._aggregate_game_dfs(game_dfs, year, day_num)

        return df_day

    @delayed
    def _get_team_roster(self, team):
        team_id = int(self._team_spell.key_to_val("TeamID", team.abbreviation.lower()))
        return {team_id: [p.player_id for p in team.roster.players]}

    # --------------------------------------------------------------
    # Returns a dict of the rosters for each team in each season
    # --------------------------------------------------------------
    def get_rosters(self):
        if self.start_date >= datetime.strptime(self._seasons.key_to_val("DayZero", self.start_date.year + 1), "%m/%d/%Y").date():
            start_season = self.start_date.year + 1
        else:
            start_season = self.start_date.year

        if self.end_date >= datetime.strptime(self._seasons.key_to_val("DayZero", self.end_date.year + 1), "%m/%d/%Y").date():
            end_season = self.end_date.year + 1
        else:
            end_season = self.end_date.year

        rosters = Util.get_source_data("rosters", path=self._helper_path, local_root=self.local_root, gcs_root=self.gcs_root)

        for season in range(start_season, end_season + 1):
            if not str(season) in rosters.keys():
                logging.info(f"Fetching teams for {season}...")
                teams = Teams(year=str(season))
                season_rosters = {}

                team_rosters_list = Parallel(n_jobs=self.n_jobs)(
                    self._get_team_roster(team) for team in tqdm(teams, desc=f"Fetching rosters for {season}"))

                for team in team_rosters_list:
                    season_rosters.update(team)

                rosters[str(season)] = season_rosters

        rosters = dict(sorted(rosters.items()))

        with open(f"{self._helper_path}MRosters.json", "w") as outfile:
            outfile.write(json.dumps(rosters))

        return rosters

    # --------------------------------------------------------------
    # A dataframe containing a list of players who did not play in certain games
    # --------------------------------------------------------------
    @property
    def absence_df(self):
        rosters = self.get_rosters()
        
        day_dfs = Parallel(n_jobs=self.n_jobs)(
            self._process_day_absences(day, games, rosters) for day, games in tqdm(self.boxscores.games.items(),
                                                                                   desc=f"Processing games from "
                                                                                        f"{self.start_date} to "
                                                                                        f"{self.end_date}"))

        df = pd.concat(day_dfs, ignore_index=True)

        df = df[['Season', 'DayNum', 'TeamID', 'Absences']]

        return df

    # --------------------------------------------------------------
    # Converts a game date into season and day_num pair
    # --------------------------------------------------------------
    def convert_game_date(self, game_date):
        day_zero = datetime.strptime(self._seasons.key_to_val("DayZero", game_date.year + 1), "%m/%d/%Y").date()

        if game_date < day_zero:
            year = game_date.year
            day_zero = datetime.strptime(self._seasons.key_to_val("DayZero", year), "%m/%d/%Y").date()
        else:
            year = game_date.year + 1

        day_num = (game_date - day_zero).days

        return year, day_num

    # --------------------------------------------------------------
    # Combines the games from a given day into a single dataframe
    # --------------------------------------------------------------
    @staticmethod
    def _aggregate_game_dfs(game_dfs, year, day_num):
        if len(game_dfs) > 0:
            df_day = pd.concat(game_dfs, ignore_index=True)

            df_day["Season"] = year
            df_day["DayNum"] = day_num

        else:
            df_day = pd.DataFrame()

        return df_day

    # --------------------------------------------------------------
    # Convenience method to retrieve the 'predict' df for the day
    # --------------------------------------------------------------
    @classmethod
    def get_bet_day_games(cls, helper_path):
        # today = cls(start_date=date.today().isoformat(), helper_path=helper_path)
        # return today.prediction_df
        df = pd.read_csv('./tmp/data/our_data/MDeleteMe.csv')
        df['WLoc'] = 'A'
        return df

    # --------------------------------------------------------------
    # Update compact and detailed results files
    # --------------------------------------------------------------
    @classmethod
    def update_results(cls, src_path, helper_path, n_jobs=-2):

        df_compact_src = pd.read_csv(f"{src_path}MRegularSeasonCompactResults_current.csv")
        df_compact_src = df_compact_src.sort_values(by=["Season", "DayNum"], ignore_index=True)
        df_detailed_src = pd.read_csv(f"{src_path}MRegularSeasonDetailedResults_current.csv")
        df_detailed_src = df_detailed_src.sort_values(by=["Season", "DayNum"], ignore_index=True)
        df_player_src = pd.read_csv(f"{src_path}MPlayerStats.csv")
        df_player_src = df_player_src.sort_values(by=["Season", "DayNum"], ignore_index=True)
        df_absence_src = pd.read_csv(f"{src_path}MAbsences.csv")
        df_absence_src = df_absence_src.sort_values(by=["Season", "DayNum"], ignore_index=True)

        seasons_lookup = MMLookup(source='seasons')

        last_day_num_compact = df_compact_src["DayNum"].iloc[-1]
        last_day_num_detailed = df_detailed_src["DayNum"].iloc[-1]
        last_day_num_player = df_player_src["DayNum"].iloc[-1]
        last_day_num_abs = df_absence_src["DayNum"].iloc[-1]

        season_compact = df_compact_src["Season"].iloc[-1]
        season_detailed = df_detailed_src["Season"].iloc[-1]
        season_player = df_player_src["Season"].iloc[-1]
        season_abs = df_absence_src["Season"].iloc[-1]

        if last_day_num_compact == last_day_num_detailed == last_day_num_player == last_day_num_abs and \
                season_compact == season_detailed == season_player == season_abs:

            last_day_num = last_day_num_compact
            season = season_compact

            if last_day_num >= seasons_lookup.key_to_val("MaxDayNum", season):
                season += 1
                last_day_num = -1

            day_zero = datetime.strptime(seasons_lookup.key_to_val("DayZero", season), "%m/%d/%Y").date()
            start_date = day_zero + timedelta(days=int(last_day_num))

            logging.info(f"Fetching data from {start_date.isoformat()} to today...")
            new_sr_data = cls(start_date=start_date.isoformat(),
                              end_date=date.today().isoformat(),
                              helper_path=helper_path,
                              n_jobs=n_jobs)

            logging.info(f"Appending new compact results to {src_path}MRegularSeasonCompactResults_current.csv...")
            df_compact_src = pd.concat([df_compact_src, new_sr_data.compact_results_df], ignore_index=True)
            logging.info(f"Appending new detailed results to {src_path}MRegularSeasonDetailedResults_current.csv...")
            df_detailed_src = pd.concat([df_detailed_src, new_sr_data.detailed_results_df], ignore_index=True)
            logging.info(f"Appending new player stats to {src_path}MPlayerStats.csv...")
            df_player_src = pd.concat([df_player_src, new_sr_data.player_stats_df], ignore_index=True)
            logging.info(f"Appending new absences to {src_path}MAbsences.csv...")
            df_absence_src = pd.concat([df_absence_src, new_sr_data.absence_df], ignore_index=True)

        else:
            if last_day_num_compact >= seasons_lookup.key_to_val("MaxDayNum", season_compact):
                season_compact += 1
                last_day_num_compact = -1

            if last_day_num_detailed >= seasons_lookup.key_to_val("MaxDayNum", season_detailed):
                season_detailed += 1
                last_day_num_detailed = -1

            if last_day_num_player >= seasons_lookup.key_to_val("MaxDayNum", season_player):
                season_player += 1
                last_day_num_player = -1

            if last_day_num_abs >= seasons_lookup.key_to_val("MaxDayNum", season_abs):
                season_abs += 1
                last_day_num_abs = -1

            day_zero_compact = datetime.strptime(seasons_lookup.key_to_val("DayZero", season_compact),
                                                 "%m/%d/%Y").date()
            day_zero_detailed = datetime.strptime(seasons_lookup.key_to_val("DayZero", season_detailed),
                                                  "%m/%d/%Y").date()
            day_zero_player = datetime.strptime(seasons_lookup.key_to_val("DayZero", season_player), "%m/%d/%Y").date()
            day_zero_abs = datetime.strptime(seasons_lookup.key_to_val("DayZero", season_abs), "%m/%d/%Y").date()

            start_date_compact = day_zero_compact + timedelta(days=int(last_day_num_compact))
            start_date_detailed = day_zero_detailed + timedelta(days=int(last_day_num_detailed))
            start_date_player = day_zero_player + timedelta(days=int(last_day_num_player))
            start_date_abs = day_zero_abs + timedelta(days=int(last_day_num_abs))

            logging.info(f"Fetching data from "
                         f"{min(start_date_detailed, start_date_compact, start_date_player, start_date_abs)}"
                         f" to today...")

            new_sr_data_compact = cls(start_date=start_date_compact.isoformat(),
                                      end_date=date.today().isoformat(),
                                      helper_path=helper_path,
                                      n_jobs=n_jobs)

            new_sr_data_detailed = cls(start_date=start_date_detailed.isoformat(),
                                       end_date=date.today().isoformat(),
                                       helper_path=helper_path,
                                       n_jobs=n_jobs)

            new_sr_data_player = cls(start_date=start_date_player.isoformat(),
                                     end_date=date.today().isoformat(),
                                     helper_path=helper_path,
                                     n_jobs=n_jobs)

            new_sr_data_abs = cls(start_date=start_date_abs.isoformat(),
                                  end_date=date.today().isoformat(),
                                  helper_path=helper_path,
                                  n_jobs=n_jobs)

            logging.info(f"Appending new compact results to {src_path}MRegularSeasonCompactResults_current.csv...")
            df_compact_src = pd.concat([df_compact_src, new_sr_data_compact.compact_results_df], ignore_index=True)
            logging.info(f"Appending new detailed results to {src_path}MRegularSeasonDetailedResults_current.csv...")
            df_detailed_src = pd.concat([df_detailed_src, new_sr_data_detailed.detailed_results_df], ignore_index=True)
            logging.info(f"Appending new player stats to {src_path}MPlayerStats.csv...")
            df_player_src = pd.concat([df_player_src, new_sr_data_player.player_stats_df], ignore_index=True)
            logging.info(f"Appending new absences to {src_path}MAbsences.csv...")
            df_absence_src = pd.concat([df_absence_src, new_sr_data_abs.absence_df], ignore_index=True)

        df_compact_src = df_compact_src.drop_duplicates(subset=["Season", "DayNum", "WTeamID", "LTeamID"])
        df_detailed_src = df_detailed_src.drop_duplicates(subset=["Season", "DayNum", "WTeamID", "LTeamID"])
        df_player_src = df_player_src.drop_duplicates(subset=["Season", "DayNum", "PlayerID"])
        df_absence_src = df_absence_src.drop_duplicates(subset=["Season", "DayNum", "Absences"])

        df_compact_src.to_csv(f"{src_path}MRegularSeasonCompactResults_current.csv", index=False)
        df_detailed_src.to_csv(f"{src_path}MRegularSeasonDetailedResults_current.csv", index=False)
        df_player_src.to_csv(f"{src_path}MPlayerStats.csv", index=False)
        df_absence_src.to_csv(f"{src_path}MAbsences.csv", index=False)


class MMOddsIngest:

    def __init__(self, day, helper_path):
        self.day = day

        self._base_url_future = f"https://api.sportsdata.io/api/cbb/"
        self._api_key = "69b201d27f95430db6dae161984008cc"
        self._headers = {"Ocp-Apim-Subscription-Key": self._api_key}

        self._api_teams = MMLookup(source='mapi_teams')
        self._seasons = MMLookup(source='seasons')

    @property
    def day_odds_df(self):
        response = requests.request("GET", f"{self._base_url_future}/odds/json/GamesByDate/{self.day}",
                                    headers=self._headers)
        data = json.loads(response.text)
        df_spreads = pd.DataFrame(data)
        df = df_spreads[["Season", "Day", "AwayTeam", "HomeTeam", "PointSpread"]]

        df = df.rename(columns={
            "Day": "DayNum",
            "AwayTeam": "WTeamID",
            "HomeTeam": "LTeamID",
            "PointSpread": "line"
        })

        df["WTeamID"] = df.apply(lambda x: self._api_teams.key_to_val("TeamID", x["WTeamID"]), axis=1)
        df["LTeamID"] = df.apply(lambda x: self._api_teams.key_to_val("TeamID", x["LTeamID"]), axis=1)

        day_zero = datetime.strptime(self._seasons.key_to_val("DayZero", df.loc[0, "Season"]), "%m/%d/%Y")
        df["DayNum"] = df.apply(lambda x: (datetime.fromisoformat(x["DayNum"]) - day_zero).days, axis=1)

        df = Util.swap_extend_spreads(df)

        return df

    def get_data_from_endpoint(self, category, endpoint):
        assert category in ["odds", "fantasy"]

        response = requests.request("GET", f"{self._base_url_future}/{category}/json/{endpoint}", headers=self._headers)
        data = json.loads(response.text)
        df = pd.DataFrame(data)

        return df, data

    # --------------------------------------------------------------
    # Update historical spread data
    # --------------------------------------------------------------
    @classmethod
    def update_spreads(cls, data_path):
        df_spreads_src = pd.read_csv(f"{data_path}MSpreads.csv")
        df_spreads_src = df_spreads_src.sort_values(by=["Season", "DayNum"], ignore_index=True)

        seasons_lookup = MMLookup(source='seasons')

        last_day_num = df_spreads_src["DayNum"].iloc[-1]
        season = df_spreads_src["Season"].iloc[-1]

        if last_day_num >= seasons_lookup.key_to_val("MaxDayNum", season):
            season += 1
        else:
            df_spreads_src = df_spreads_src[df_spreads_src["Season"] < season]

        game_date = date.today()
        day_zero = datetime.strptime(seasons_lookup.key_to_val("DayZero", game_date.year + 1), "%m/%d/%Y").date()

        if game_date < day_zero:
            current_season = game_date.year
        else:
            current_season = game_date.year + 1

        season_dfs = []

        if season > current_season:
            logging.info("Spreads already up to game_date")

        else:
            logging.info(f'Collecting spread data from {season} to {current_season}...')
            for y in range(season, current_season + 1):
                url = f"https://www.thepredictiontracker.com/ncaabb{str(y - 1)[-2:]}.csv"
                req = Request(url)
                req.add_header('User-Agent',
                               'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:77.0) Gecko/20100101 Firefox/77.0')
                content = urlopen(req)
                df_season = pd.read_csv(content, parse_dates=["date"], infer_datetime_format=True, na_values=".")

                df_season["Season"] = y
                day_zero = datetime.strptime(seasons_lookup.key_to_val("DayZero", y), "%m/%d/%Y")
                df_season["DayNum"] = df_season.apply(lambda x: (x["date"] - day_zero).days, axis=1)
                season_dfs += [df_season]

                logging.info(f'Completed {y}...')

            new_spreads_df = pd.concat(season_dfs)

            logging.info('Converting team names to ids and converting to win/loss "kaggle" format...')
            team_lookup = MMLookup(source='team_spellings')
            new_spreads_df = new_spreads_df[new_spreads_df["home"].notna() & new_spreads_df["road"].notna()]
            new_spreads_df["homeID"] = new_spreads_df.apply(lambda x: team_lookup.key_to_val("TeamID", x.home.lower()),
                                                            axis=1)
            new_spreads_df["roadID"] = new_spreads_df.apply(lambda x: team_lookup.key_to_val("TeamID", x.road.lower()),
                                                            axis=1)

            df_home = new_spreads_df[new_spreads_df["hscore"] > new_spreads_df["rscore"]]
            df_home = df_home.rename(columns={"homeID": "WTeamID",
                                              "hscore": "WScore",
                                              "roadID": "LTeamID",
                                              "rscore": "LScore"})

            df_road = new_spreads_df[new_spreads_df["hscore"] < new_spreads_df["rscore"]]
            df_road = df_road.rename(columns={"homeID": "LTeamID",
                                              "hscore": "LScore",
                                              "roadID": "WTeamID",
                                              "rscore": "WScore"})
            df_road["line"] = -df_road["line"]

            kag_cols = ["Season", "DayNum", "WTeamID", "WScore", "LTeamID", "LScore", "line"]
            new_spreads_df = pd.concat([df_road[kag_cols], df_home[kag_cols]], ignore_index=True)

            new_spreads_df = new_spreads_df.dropna(subset=["DayNum"])
            new_spreads_df = new_spreads_df.astype({"DayNum": int})
            new_spreads_df = new_spreads_df.sort_values(by=["Season", "DayNum"], ignore_index=True)

            df_spreads_src = df_spreads_src.append(new_spreads_df)
            df_spreads_src = df_spreads_src.drop_duplicates(subset=["Season", "DayNum", "WTeamID", "LTeamID"])

            logging.info(f'Saving updated spreads to {data_path}MSpreads.csv...')
            df_spreads_src.to_csv(f"{data_path}MSpreads.csv", index=False)

    @classmethod
    def get_bet_day_spreads_df(cls, helper_path):
        df = pd.read_csv('./tmp/data/our_data/MDeleteMe2.csv')
        return df
        # return cls(day=date.today().isoformat(), helper_path=helper_path).day_odds_df


class InjuryIngestor:
    def __init__(self, save_path, helper_path="./data/our_data/", remove_file=True):
        self.save_path = save_path
        self.remove_file = remove_file

        self._injury_url = "https://www.rotowire.com/cbasketball/injury-report.php"
        self._options = webdriver.ChromeOptions()
        self._options.add_argument("--headless")
        prefs = {"download.default_directory": os.path.abspath(save_path)}
        self._options.add_experimental_option("prefs", prefs)

        self._team_spell = MMLookup(source='team_spellings')

        day = date.today()
        sr_ingest = MMSRIngest(day.isoformat(), helper_path=helper_path)
        self._season, self._day_num = sr_ingest.convert_game_date(day)
        self._rosters = sr_ingest.get_rosters()

        self._odd_names = {
            "RJ Gunn": "rodney-gunnjr-1",
            "Bobby Pettiford": "bob-pettiford-1"
        }

    def _download_injuries(self):
        downloaded = False
        browser = webdriver.Chrome(service=Service(ChromeDriverManager().install()), options=self._options)

        browser.get(self._injury_url)

        try:
            download_button = WebDriverWait(browser, 10).until(
                expected_conditions.presence_of_element_located((By.XPATH, "//button[@class='export-button is-csv']")))
            download_button.click()
            wait_count = 0
            download_path = f"{self.save_path}college-basketball-injury-report.csv"

            while wait_count < 50 and not downloaded:
                time.sleep(0.1)
                if os.path.exists(download_path):
                    downloaded = True
                else:
                    wait_count += 1

            if downloaded:
                logging.info("Injury file successfully downloaded")
            else:
                raise Exception({"message": "file was not successfully downloaded"})

        finally:
            browser.quit()

        return downloaded

    def get_today_injuries(self):
        assert self._download_injuries()

        df = pd.read_csv(f"{self.save_path}college-basketball-injury-report.csv")
        df = df[["Player", "Team", "Status"]]
        df = df.rename(columns={
            "Team": "TeamID"
        })

        df["TeamID"] = df.apply(lambda x: int(self._team_spell.key_to_val("TeamID", x["TeamID"].lower())), axis=1)
        df = df[df["TeamID"] != -1]
        df["PlayerID"] = df.apply(lambda x: self._get_player_id(x), axis=1)

        if self.remove_file:
            os.remove(f"{self.save_path}college-basketball-injury-report.csv")

        return df

    def _get_player_id(self, row):
        correct_id = None

        if row["Player"] in self._odd_names.keys():
            correct_id = self._odd_names[row["Player"]]

        else:
            team_roster = self._rosters[str(self._season)][str(row["TeamID"])]
            player_names = row["Player"].split(" ")
            player_first = "".join(l for l in player_names[0] if l.isalnum() or l == "-").lower()
            player_last = "".join(l for l in player_names[-1] if l.isalnum() or l == "-").lower()
            if player_last in ["jr", "sr", "ii", "iii", "iv"]:
                player_last = "".join(l for l in player_names[-2] if l.isalnum() or l == "-").lower()
            elif len(player_names) > 2:
                player_last = "".join(l for l in player_names[-2] if l.isalnum() or l == "-").lower() + player_last

            player_name = f"{player_first} {player_last}"

            for pid in team_roster:
                id_names = pid.split("-")

                id_first = id_names[0]
                if len(id_names) > 3:
                    id_last = f"{id_names[-3]}-{id_names[-2]}"
                else:
                    id_last = id_names[-2]

                if id_last[-2:] in ["jr", "sr", "ii", "iv"]:
                    id_last = id_last[:-2]
                elif id_last[-3:] == "iii":
                    id_last = id_last[:-3]

                id_name = f"{id_first} {id_last}"

                if id_name.startswith(player_name) or player_name.startswith(id_name):
                    correct_id = pid
                    break

        if correct_id is None:
            return -1
        else:
            return correct_id
