# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import logging
import re

import numpy as np
import pandas as pd

import src.mmutil as mu


# -----------------------------------------------------------------------------
#
#   Name:      MMMetric
#
#   Purpose:   Base class for all metrics.  Basically does nothing except open
#              a source df.  Subclass and modify as appropriate.
#
#   Usage:     m = MMMetric(start_yr=2003,
#                           end_yr=2020,
#                           calc_field='Score',
#                           metric_name='Raw_Seed',
#                           feature_name='Seed',
#                           data_path=data_path)
#              df = m.get_metric()
#              TODO: Document parameters
#
# -----------------------------------------------------------------------------
class MMMetric:
    def __init__(self, *args, **kwargs):
        self.metric_name = kwargs.pop('metric_name')  # "Spread"
        self.feature_name = kwargs.pop('feature_name')  # "Team_Q" <- Spread~-1+T1_TeamID+T2_TeamID
        self.path = kwargs.pop('path')
        self.source = kwargs.pop('source')
        self.segment_count = kwargs.pop('segment_count')
        self.start = kwargs.pop('start')
        self.end = kwargs.pop('end')
        self.gcs_root = kwargs.pop('gcs_root') if 'gcs_root' in kwargs.keys() else None
        self.local_root = kwargs.pop('local_root') if 'local_root' in kwargs.keys() else None
        self._set_df_source()

    def _set_df_source(self):
        if isinstance(self.source, pd.DataFrame):
            # here we assume if we're passed a df its already cropped and segmented
            self.df_source = self.source
        else:
            # Open, crop, and segment
            self.df_source = mu.Util.get_source_data(self.source, self.path, gcs_root=self.gcs_root, local_root=self.local_root)

            if not self.df_source.empty:
                self.df_source = self.df_source.sort_values(by=["Season", "DayNum"], ignore_index=True)
                self.df_source = self.df_source.reset_index(drop=True)
                self.df_source = mu.Util.chunk_seasons(self.df_source, segment_count=self.segment_count)
            else:
                logging.debug(f'No source df for metric: {self.metric_name}...')

    # ------------------------------------------------------------------------
    # Gets a metric from which a feature_def may be built
    # ------------------------------------------------------------------------
    def get_metric(self):
        # calculate
        df = self.df_source.copy()
        df = self._calc_metric(df)
        df = self._conform_output(df)
        return df

    # ------------------------------------------------------------------------
    # Empty support function to be replaced by subclasses
    # ------------------------------------------------------------------------
    def _calc_metric(self, df):
        return df

    # ------------------------------------------------------------------------
    # Empty support function to be replaced by subclasses
    # ------------------------------------------------------------------------
    def _conform_output(self, df):
        return df


# -----------------------------------------------------------------------------
#
#   Name:      MMMetricSeed(MMMetric)
#
#   Purpose:   Example use of MMMetric to use the Seed file to be passed to
#              MMFeature.
#
#   Usage:     m = MMMetricSeed(start_yr=2003,
#                               end_yr=2020,
#                               calc_field='Score',
#                               feature_name='Seed',
#                               data_path=data_path)
#              df = m.get_metric()
#              TODO: Document parameters
#
# -----------------------------------------------------------------------------
class MMMetricSeed(MMMetric):
    def _conform_output(self, df):

        try:
            df[self.feature_name] = df[self.feature_name].apply(lambda x: int(re.sub("[^0-9]", " ", x)))
        except TypeError:
            pass

        return df


# -----------------------------------------------------------------------------
#
#   Name:      MMMetricGame(MMMetric)
#
#   Purpose:   This provides game level metrics of the form:
#              df['Season', 'T1_TeamID', 'T2_TeamID', 'My_Metric'].
#              The df returned here is a source for MMFeatureGLM
#
#   Usage:     m = MMMetricSeed(start_yr=2003,
#                               end_yr=2020,
#                               calc_field='Score',
#                               feature_name='Team_Q',
#                               metric_name='Spread'
#                               data_path=data_path)
#              df = m.get_metric()
#              TODO: Document parameters
#
# -----------------------------------------------------------------------------
class MMMetricGame(MMMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.op = kwargs.pop('op')
        self.lag = kwargs.pop('lag')
        self.window = kwargs.pop('window')

        # Adjust start and end for window and lag
        # self.start, self.end = mu.Util.adjust_for_lag_and_window(start=self.start, end=self.end, lag=self.lag,
        #                                                          window=self.window, segment_count=self.segment_count)
        self.df_source = mu.Util.crop(self.df_source, start=self.start, end=self.end)

    def get_metric(self):
        df = self.df_source
        df = mu.Util.swap_extend(df)
        df = self._calc_metric(df)
        df = self._conform_output(df)

        return df

    def _calc_metric(self, df):
        df[self.metric_name] = self.op(df)
        df = df[['Season', 'Segment', 'DayNum', 'T1_TeamID', 'T2_TeamID', self.metric_name]]

        return df

    def _conform_output(self, df):
        df = df.rename(columns={self.metric_name: self.feature_name})

        return df


# -----------------------------------------------------------------------------
#
#   Name:      MMMetricSpread(MMMetric)
#
# -----------------------------------------------------------------------------
class MMMetricSpread(MMMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.op = kwargs.pop('op')

    def get_metric(self):
        df = self.df_source
        df = mu.Util.swap_extend_spreads(df)
        df = self._calc_metric(df)
        df = self._conform_output(df)

        return df

    def _calc_metric(self, df):
        df[self.metric_name] = self.op(df)
        df = df[['Season', 'Segment', 'DayNum', 'T1_TeamID', 'T2_TeamID', self.metric_name]]

        return df

    def _conform_output(self, df):
        df = df.rename(columns={self.metric_name: self.feature_name})

        return df


class MMMetricPred(MMMetric):

    def get_metric(self):
        # TODO: preds should already be swapped, so I don't think we need a swapper here
        df = self.df_source
        df = self._calc_metric(df)
        df = self._conform_output(df)

        return df

    def _calc_metric(self, df):
        df = df[['Season', 'Segment', 'DayNum', 'T1_TeamID', 'T2_TeamID', self.metric_name]]

        return df

    def _conform_output(self, df):
        df = df.rename(columns={self.metric_name: self.feature_name})

        return df


# -----------------------------------------------------------------------------
#
#   Name:      MMMetricSeason(MMMetric)
#
#   Purpose:   This provides season level metrics of the form:
#              df['Season', 'T1_TeamID', 'T2_TeamID', 'My_Metric'].
#              The df returned here is a source for MMFeatureGLM
#
#   Usage:     m = MMMetricSeed(start_yr=2003,
#                               end_yr=2020,
#                               calc_field='Score',
#                               feature_name='Team_Q',
#                               metric_name='Spread'
#                               data_path=data_path)
#              df = m.get_metric()
#              TODO: Document parameters
#
# -----------------------------------------------------------------------------
class MMMetricSeason(MMMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.op = kwargs.pop('op')
        self.aggregation = kwargs.pop("aggregation")
        self.lag = kwargs.pop('lag')
        self.window = kwargs.pop('window')

        # Adjust start and end for window and lag
        self.adj_start, adj_end = mu.Util.adjust_for_lag_and_window(start=self.start, end=self.end, lag=self.lag,
                                                                    window=self.window,
                                                                    segment_count=self.segment_count)
        self.df_source = mu.Util.crop(self.df_source, start=self.adj_start, end=adj_end)

    def get_metric(self):
        if "T1_TeamID" in self.df_source.columns:
            df = self.df_source.copy()

        else:
            df = mu.Util.swap_extend(self.df_source)

        df = self._calc_metric(df)
        df = self._conform_output(df)
        return df

    def _calc_metric(self, df):

        tot_segs = ((self.end[0] - self.start[0]) * self.segment_count) + (self.end[1] - self.start[1])
        window_start, window_end = mu.Util.get_window(start=self.adj_start, window=self.window,
                                                      segment_count=self.segment_count)

        segments = []

        for seg in range(tot_segs + 1):
            if self.aggregation == "sum":
                df_seg = mu.Util.crop(df, window_start, window_end).groupby(["T1_TeamID"]).sum().reset_index()
            elif self.aggregation == "mean":
                df_seg = mu.Util.crop(df, window_start, window_end).groupby(["T1_TeamID"]).mean().reset_index()
            else:
                raise ValueError("Invalid aggregation method for MMMetricSeason.")

            df_seg[self.metric_name] = self.op(df_seg)

            metric_end = mu.Util.adjust_date(window_end, adj=self.lag, segment_count=self.segment_count)
            df_seg["Season"] = metric_end[0]
            df_seg["Segment"] = metric_end[1]
            df_seg = df_seg[['Season', 'Segment', 'T1_TeamID', self.metric_name]]
            segments += [df_seg]

            new_start_seg = (window_start[1] + 1) % self.segment_count
            new_start_year = window_start[0] + (window_start[1] + 1) // self.segment_count
            window_start, window_end = mu.Util.get_window(start=(new_start_year, new_start_seg, self.start[2]),
                                                          window=self.window, segment_count=self.segment_count)

        df = pd.concat(segments, ignore_index=True)

        return df

    def _conform_output(self, df):
        df = df.rename(columns={
            "T1_TeamID": "TeamID",
            self.metric_name: self.feature_name
        })
        return df


class MMMetricAbsence(MMMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.op = kwargs.pop("op")

        self.df_player_stats = mu.Util.get_source_data("player_stats", path=self.path, gcs_root=self.gcs_root, local_root=self.local_root)
        self.df_player_stats[self.metric_name] = self.op(self.df_player_stats)
        self.df_player_stats = self.df_player_stats[["Season", "DayNum", "PlayerID", self.metric_name]]

    def _set_df_source(self):
        if isinstance(self.source, pd.DataFrame):
            # here we assume if we're passed a df its already cropped and segmented
            self.df_source = self.source
        else:
            # Open and segment
            self.df_source = mu.Util.get_source_data(self.source, path=self.path, gcs_root=self.gcs_root, local_root=self.local_root)
            self.df_source = self.df_source.sort_values(by=["Season", "DayNum"], ignore_index=True)
            self.df_source = self.df_source.reset_index(drop=True)

    def get_metric(self):
        df = self.df_source
        df = self._calc_metric(df)

        return df

    def _get_absent_stat(self, season):
        df_season_stats = self.df_player_stats[self.df_player_stats["Season"] == season].drop(columns=["Season"])
        days = df_season_stats["DayNum"].unique()
        day_dfs = []

        for day in days:
            df_day_stats = df_season_stats[df_season_stats["DayNum"] < day].drop(columns=["DayNum"])
            df_day_stats = df_day_stats.groupby(["PlayerID"], as_index=False).mean()
            df_day_stats["DayNum"] = day
            day_dfs.append(df_day_stats)

        if len(day_dfs) > 0:
            df_season = pd.concat(day_dfs, ignore_index=True)
            df_season["Season"] = season
        else:
            df_season = pd.DataFrame()

        return df_season

    def _calc_metric(self, df):
        season_dfs = [self._get_absent_stat(season) for season in range(self.start[0], self.end[0] + 1)]

        df_stats = pd.concat(season_dfs, ignore_index=True)
        if not df_stats.empty:
            df = df.merge(df_stats, left_on=["Season", "DayNum", "Absences"], right_on=["Season", "DayNum", "PlayerID"])
            df = df.groupby(["Season", "DayNum", "TeamID"], as_index=False).sum()
        else:
            df = pd.DataFrame()

        return df


# -----------------------------------------------------------------------------
#
#   Name:      MMTeamTransform()
#
#   Purpose:   Function holder for metric and target ops
#
#   Usage:     MMTeamTransform.spread(df)
#              TODO: Document parameters
#
# -----------------------------------------------------------------------------
class MMTeamTransform:

    # -----------------------------------------------------------------------------
    # Turns score to win/loss win=1, loss=0
    # Formula: NA
    # -----------------------------------------------------------------------------
    @classmethod
    def win(cls, df):
        col = pd.Series(data=0, index=df.index)
        col[df['T1_Score'] > df['T2_Score']] = 1

        return col
        # return 1 if df['T1_Score'] > df['T2_Score'] else 0

    # -----------------------------------------------------------------------------
    # Point differential
    # Formula: T1_Score - T2_Score
    # -----------------------------------------------------------------------------
    @classmethod
    def mov(cls, df): return df['T1_Score'] - df['T2_Score']

    # Vegas spread
    @classmethod
    def spread(cls, df): return df["line"]

    # Which side covered the spread
    # 0: push
    # 1: T1 covered
    # 2: T2 covered
    @classmethod
    def cover(cls, df):
        col = pd.Series(data=np.nan, index=df.index)
        col[df['T1_Score'] - df['T2_Score'] == df['line']] = 0
        col[df['T1_Score'] - df['T2_Score'] > df['line']] = 1
        col[df['T1_Score'] - df['T2_Score'] < df['line']] = 2

        return col

        # if pd.isna(df['line']):
        #     return np.nan
        # if df['T1_Score'] - df['T2_Score'] > df['line']:
        #     return 1
        # elif df['T1_Score'] - df['T2_Score'] == df['line']:
        #     return 0
        # else:
        #     return 2

    @classmethod
    def payout(cls, df):
        col = pd.Series(data=np.nan, index=df.index)
        col[df['T1_Score'] - df['T2_Score'] == df['line']] = 1
        col[df['T1_Score'] - df['T2_Score'] > df['line']] = 2.1 / 1.1
        col[df['T1_Score'] - df['T2_Score'] < df['line']] = 0

        return col

    # Location: Home, Away, Neutral
    @classmethod
    def location(cls, df):
        return df["location"]

    # T1 total possessions
    # Formula: 0.96 * (FGA + TO + 0.475 * FTA)
    @classmethod
    def possessions(cls, df): return 0.96 * (df['T1_FGA'] + df['T1_TO'] + (0.475 * df['T1_FTA']))

    # Half total possessions
    # Formula: (possessions(T1) + possessions(T2))/2
    @classmethod
    def half_tot_possessions(cls, df):
        return 0.96 * (df['T1_FGA'] + df["T2_FGA"] + df['T1_TO'] + df['T2_TO'] +
                       (0.475 * (df['T1_FTA'] + df['T2_FTA']))) / 2

    # Half total possessions adjusted for overtimes
    # Formula: half_tot_possessions / (1 + 0.125 * NumOT)
    @classmethod
    def adj_possessions(cls, df):
        return cls.half_tot_possessions(df) / (1 + 0.125 * df["NumOT"])

    # Points scored by T1
    @classmethod
    def off_score(cls, df): return df["T1_Score"]

    # Points scored by T2
    @classmethod
    def def_score(cls, df): return df["T2_Score"]

    # Points per possession scored by T1
    # Formula: off_score / half_tot_possessions
    @classmethod
    def off_rtg(cls, df): return df["T1_Score"] / cls.half_tot_possessions(df)

    # Points per possession scored by T2
    # Formula: def_score / half_tot_possessions
    @classmethod
    def def_rtg(cls, df): return df["T2_Score"] / cls.half_tot_possessions(df)

    # Hollinger assist ratio. Ratio of field goals on which there was an assist.
    # Formula: Ast / (FGA + 0.475 * FTA + Ast + TO)
    @classmethod
    def assist_ratio(cls, df): return df["T1_Ast"] / (df["T1_FGA"] + 0.475 * df["T1_FTA"] + df["T1_Ast"] +
                                                      df["T1_TO"])

    # Estimate of turnovers per play
    # Formula: TO / (FGA + 0.475 * FTA + Ast + TO)
    @classmethod
    def turnover_ratio(cls, df): return df["T1_TO"] / (df["T1_FGA"] + 0.475 * df["T1_FTA"] + df["T1_Ast"] +
                                                       df["T1_TO"])

    # A better estimate of a teams shooting percentage, taking FGs and FTs into account
    # Formula: Score / (2 * (FGA + 0.475 * FTA))
    @classmethod
    def true_shooting(cls, df): return df["T1_Score"] / (2 * (df["T1_FGA"] + 0.475 * df["T1_FTA"]))

    # A better estimate of a teams FG%, taking the higher value of 3 point FGs into account
    # Formula: (FGM + 0.5 * FGM3) / FGA
    @classmethod
    def effective_fg_pct(cls, df): return (df["T1_FGM"] + 0.5 * df["T1_FGM3"]) / df["T1_FGA"]

    # Offensive rebounds made as a percentage of available offensive rebounds
    # Formula: T1_OR / (T1_OR + T2_DR)
    @classmethod
    def off_reb_pct(cls, df): return df["T1_OR"] / (df["T1_OR"] + df["T2_DR"])

    # Defensive rebounds made as a percentage of available defensive rebounds
    # Formula: T1_DR / (T1_DR + T2_OR)
    @classmethod
    def def_reb_pct(cls, df): return df["T1_DR"] / (df["T1_DR"] + df["T2_OR"])

    # Rebounds made as a percent of available rebounds
    # Formula: (T1_OR + T1_DR) / (T1_OR + T2_DR + T1_DR + T2_OR)
    @classmethod
    def tot_reb_pct(cls, df): return (df["T1_OR"] + df["T1_DR"]) / \
                                     (df["T1_OR"] + df["T2_DR"] + df["T1_DR"] + df["T2_OR"])

    # A measure of how often a team gets to the FT line
    # Formula: FTA / FGA
    @classmethod
    def ft_rate(cls, df): return df["T1_FTA"] / df["T1_FGA"]

    # Strength of schedule. This is a "2nd degree" metric which measures the quality of teams played. Requires Team_Q
    # to have already been calculated
    @classmethod
    def sos(cls, df): return df["T2_Team_Q"]


class MMPlayerTransform:
    @classmethod
    def points(cls, df): return df["Pt"]

    @classmethod
    def off_reb(cls, df): return df["OR"]

    @classmethod
    def def_reb(cls, df): return df["DR"]

    @classmethod
    def assists(cls, df): return df["Ast"]

    @classmethod
    def turnovers(cls, df): return df["TO"]

    @classmethod
    def blocks(cls, df): return df["Blk"]

    @classmethod
    def steals(cls, df): return df["Stl"]

    @classmethod
    def fouls(cls, df): return df["PF"]

    @classmethod
    def minutes(cls, df): return df["MP"]

    @classmethod
    def effective_fg_pct(cls, df): return (df["FGM"] + 0.5 * df["FGM3"]) / df["FGA"]

    @classmethod
    def true_shooting(cls, df): return df["Score"] / (2 * (df["FGA"] + 0.475 * df["FTA"]))

    # Player Impact Estimator
    # Formula: Pt + FGM + FTM - FGA - FTA + DR + 0.5 * OR + Ast + Stl + 0.5 * Blk - PF - TO
    @classmethod
    def pie(cls, df): return df["Pt"] + df["FGM"] + df["FTM"] - df["FGA"] - df["FTA"] + df["DR"] + 0.5 * df["OR"] + \
                             df["Ast"] + df["Stl"] + 0.5 * df["Blk"] - df["PF"] - df["TO"]
