# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import logging

from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, PolynomialFeatures

import src.mmdata as md
import src.mmfeature as mf
import src.mmmetric as mm
from src.mmdata import C_COLS
from src.mmingest import MMSRIngest
from src.mmutil import Util


# ----------------------------------------------------------------------------------------------------------------------
#
#   Name:       MMSessionBase
#
#   Purpose:    Base class for controlling object for an MM training, or prediction session
#               most parameters for both MMTrain and MMData are set here.  Not to be used directly, but
#               instead, used to inherit sessions from.
#
#
#   Usage:      features = ['Team_Q', 'Off_Q', 'Def_Q', 'Off_Poss_Q', 'Def_Poss_Q']
#               feature_list = []
#
#               for feature_def in features:
#                   feature_list.append(mf.MMFeatureDefinition.get_feature_def(feature_def))
#
#               session_params = {
#                   'feature_list': feature_list,
#                   'path': '../',
#                   'segment_count': 4,
#                   'train_start': (2004, 1),
#                   'train_end': (2021, 2),
#                   'test_start': (2021, 3),
#                   'test_end': (2022, 2),
#                   'future': False
# }
#
# config = ms.MMSessionMOVCurrentSeason(session_params)
#
# ----------------------------------------------------------------------------------------------------------------------

class MMSessionBase:

    def __init__(self, session_params):

        self.local_root = session_params['local_root']
        self.gcs_root = session_params['gcs_root']
        self.path = session_params['path']
        self.future = session_params['future']
        self.log_level = session_params['log_level']        # this is the logging level overall

        # Paths
        self.data_path = self.path + 'data/our_data/'       # top level data_path for the project
        self.feature_path = self.path + 'features/'         # place to store/get computed features
        self.model_path = self.path + 'models/'             # path to store models after training
        self.preds_path = self.path + 'output/preds/'       # path to store predictions
        self.submit_path = self.path + 'output/submit/'     # path to store submission

        # Data operation settings
        self.refresh = True                                 # True to rebuild but use prebuilt features, False to use last built datasets
        self.csv_name = 'MMLast.csv'                        # Root file name for saved train/test/val sets
        self.to_h2o = True                                  # True for an H2OFrame, False for a df
        self.njobs = -2                                     # -2 to parallelize data build, 1 for debugging/single threaded

        # Scikit-learn pipeline for end of data pipe
        self.preprocessor = Pipeline([("scaler", StandardScaler()),
                                      ('poly', PolynomialFeatures(degree=3, include_bias=False))])

        # Training and reporting cols
        self.train_cols = C_COLS                        # Columns to exclude/include in the data
        self.exclude = True                             # True if train_cols are exclusions, False to include

        # aml session settings
        self.sort_metric = 'AUTO'                       # sort order for the Leaderboard
        self.nfolds = 5                                 # number of folds for k folds cross val training
        self.seed = 42                                  # Seed to feed for reproducible results
        self.stopping_metric = 'AUTO'                   # See aml docs for other options
        self.stopping_rounds = 3                        # num of epochs to stop when stopping metric doesn't improve
        self.exclude_algos = []                         # algos to exclude from aml, see aml docs.  Must exclude DeepLearning for reproducibility
        self.max_runtime_secs = 0                       # runtime for aml in secs, 0 for unlimited, cannot be set for reproducibility
        self.max_models = 15                            # max number of models, must be set for reproducibility
        self.max_runtime_secs_per_model = 0             # means what it says
        self.include_algos = None                       # algos to include, must be None, if exclude_algos is not None
        self.exploitation_ratio = 0                     # can experiment with a ratio of 0.1 -> see docs
        self.keep_cross_validation_predictions = False  # really only matters for multiple runs on same Leaderboard
        self.keep_cross_validation_fold_assignment = False
        self.export_checkpoints_dir = None
        self.balance_classes = False                    # built-in parameter to deal with class imbalance



        # in debug mode, h2o is also in debug mode
        self.verbosity = 'Debug' if self.log_level == 'debug' else None
        logging.getLogger().setLevel(self.log_level)

        # Settings set and managed internally to MMSessionConfig/MMTrain/MMData
        self.to_csv = self.refresh                      # Save features when they are refreshed
        self.run_id = ''                                # ID for a session
        self.step_num = 0                               # Step number within a session
        self.lb_master = ''                             # Master leaderboard with run_id and step added
        self.aml_master = []                            # List of aml objects used in this session

        # make sure we have good paths everywhere
        Util.make_paths(path_list=[self.data_path, self.model_path, self.preds_path, self.feature_path,
                                   self.feature_path + 'data', self.submit_path])

        # set session params to None, all child classes will need to fill these in by overriding self.set_session params()
        self.feature_list = None                                # A list of MMFeatures, see MMFeatureDefinition
        self.path = None                                        # top level path for the project
        self.source = None                                      # source for df_core (both train and test if test_source is None
        self.test_source = None                                 # source for df_core_test if not None
        self.train_target_source = None                         # source to get training targets from
        self.test_target_source = None                          # source to get test targets from
        self.segment_count = None                               # number of segments to divide seasons into
        self.train_start = None                                 # Tuple of the form [Season, Segment] for the first day of the training session
        self.train_end = None                                   # Same as above for the last day of the training session
        self.test_start = None                                  # Tuple of the form [Season, Segment] for the first day of the test set
        self.test_end = None                                    # Same as above for the last day of the test set
        self.future = None                                      # Must be true if the gt is unknown
        self.description = None                                 # Brief description of features used, only used in model_db

        self.set_session_params(session_params)                 # load session params

        self.feature_cols = []                                  # List of all feature columns converted to T1 T2 format
        self.categorical_cols = []                              # List of categorical columns
        self.numeric_cols = []                                  # List of numeric columns
        self.set_feature_cols()                                 # set the values of feature_cols, categorical_cols, and numeric_cols

        # Load prediction file if future is true
        if self.future:
            self.df_pred = MMSRIngest.get_bet_day_games(self.data_path)
        else:
            self.df_pred = None

        # Load data object
        self.data = md.MMData(self)  # the current session's data object

    def set_session_params(self, session_params):

        self.feature_list = session_params['feature_list']      # See init for definitions
        self.path = session_params['path']
        self.local_root = session_params['local_root']
        self.gcs_root = session_params['gcs_root']
        self.source = session_params['source']
        self.test_source = session_params['test_source']
        self.train_target_source = session_params['train_target_source']
        self.test_target_source = session_params['test_target_source']
        self.segment_count = session_params['segment_count']
        self.train_start = session_params['train_start']
        self.train_end = session_params['train_end']
        self.test_start = session_params['test_start']
        self.test_end = session_params['test_end']
        self.future = session_params['future']
        self.log_level = session_params['log_level']
        self.description = session_params['description']

    def delete_features(self):
        for feature in self.feature_list:
            if feature["feature_obj"] in [mf.MMFeatureOrdinal, mf.MMFeatureAbsence]:
                segment_count = None

            else:
                segment_count = self.segment_count

            if "lag" in feature["feature_params"].keys() and "window" in feature["feature_params"].keys():
                Util.delete_saved_feature(self.feature_path,
                                          feature["feature_params"]["feature_name"],
                                          segment_count=segment_count,
                                          lag=feature["feature_params"]["lag"],
                                          window=feature["feature_params"]["window"])

            elif "window" in feature["feature_params"].keys():
                Util.delete_saved_feature(self.feature_path,
                                          feature["feature_params"]["feature_name"],
                                          segment_count=segment_count,
                                          window=feature["feature_params"]["window"])

            else:
                Util.delete_saved_feature(self.feature_path,
                                          feature["feature_params"]["feature_name"],
                                          segment_count=segment_count)

    def set_feature_cols(self):
        for feature in self.feature_list:
            if feature["feature_obj"] == mf.MMFeatureKeyVal:
                self.feature_cols.append(feature["feature_params"]["feature_name"])

            elif feature["feature_obj"] == mf.MMFeatureCategorical:
                self.feature_cols.append(feature["feature_params"]["feature_name"])
                self.categorical_cols.append(feature["feature_params"]["feature_name"])

            else:
                self.feature_cols.extend([f'T1_{feature["feature_params"]["feature_name"]}',
                                          f'T2_{feature["feature_params"]["feature_name"]}'])

        self.numeric_cols = list(set(self.feature_cols).difference(self.categorical_cols))

    def extend_report_cols(self):
        self.report_cols.extend(self.feature_cols)


# ----------------------------------------------------------------------------------------------------------------------
#
#   Name:       MMSessionMOVCurrentSeasonBase
#
#   Purpose:    Config for predicting MOV using the Current Season ('cur_season') as your data source. See base class
#               for more detail.
#
#   Usage:      config = ms.MMSessionMOVCurrentSeason(session_params)
#
# ----------------------------------------------------------------------------------------------------------------------

class MMSessionMOVCurrentSeason(MMSessionBase):

    def __init__(self, session_params):

        super().__init__(session_params)

        # target (predicted y) column settings
        self.target_col = 'MOV'                             # Column name of the predicted y
        self.target_type = 'continuous'                     # Tells if y is  'binary' or 'continuous'
        self.target_op = mm.MMTeamTransform.mov             # MMTeamTransform to calculate the target col
        self.exclude_algos = ['DRF']                        # exclude DRF for continuous variables
        self.report_cols = ['Season',                       # Col names to add to predict set for post reporting (other than target_col)
                            'Segment',
                            'DayNum',
                            'T1_TeamID',
                            'T2_TeamID',
                            self.target_col]

        self.preprocessor = Pipeline([("scaler", StandardScaler()),
                                      ('poly', PolynomialFeatures(degree=2, include_bias=False))])
        self.njobs = -2

# ----------------------------------------------------------------------------------------------------------------------
#
#   Name:       MMSessionMOVCurrentSeasonBase
#
#   Purpose:    Config for predicting MOV using the Current Season ('cur_season') as your data source. See base class
#               for more detail.
#
#   Usage:      config = ms.MMSessionMOVCurrentSeason(session_params)
#
# ----------------------------------------------------------------------------------------------------------------------

class MMSessionMOVMarchMad(MMSessionBase):

    def __init__(self, session_params):

        self.test_source = 'march_mad'  # TODO: does this need to come before the super below?

        super().__init__(session_params)

        # target (predicted y) column settings
        self.target_col = 'MOV'                             # Column name of the predicted y
        self.train_target_source = 'cur_season'             # Source of target col for training
        self.target_type = 'continuous'                     # Tells if y is  'binary' or 'continuous'
        self.target_op = mm.MMTeamTransform.mov             # MMTeamTransform to calculate the target col
        self.exclude_algos = ['DRF']                        # exclude DRF for continuous variables
        self.report_cols = ['Season',                       # Col names to add to predict set for post reporting (other than target_col)
                            'Segment',
                            'DayNum',
                            'T1_TeamID',
                            'T2_TeamID',
                            self.target_col]


# ----------------------------------------------------------------------------------------------------------------------
#
#   Name:       MMSessionPVSpreadCurrentSeasonBase
#
#   Purpose:    Config for predicting 'pVSpread' using the Current Season ('cur_season') as your data source. See base class
#               for more detail.
#
#   Usage:      config = ms.MMSessionMOVCurrentSeason(session_params)
#
# ----------------------------------------------------------------------------------------------------------------------

class MMSessionPVSpreadCurrentSeason(MMSessionBase):

    def __init__(self, session_params):
        super().__init__(session_params)

        # target (predicted y) column settings
        self.target_col = 'pVSpread'                     # Column name of the predicted y
        self.train_target_source = 'spreads'             # Source of target col for train
        self.test_target_source = 'spreads'              # Source of target col for test
        self.target_type = 'continuous'                  # Tells if y is  'binary' or 'continuous'
        self.exclude_algos = ['DRF']                     # DRF is expensive for continuous variable
        self.target_op = mm.MMTeamTransform.spread       # MMTeamTransform to calculate the target col
        self.report_cols = ['Season',                    # Col names to add to predict set for post reporting (other than target_col)
                            'Segment',
                            'DayNum',
                            'T1_TeamID',
                            'T2_TeamID',
                            self.target_col]


class MMSessionCoverMarchMad(MMSessionBase):

    def __init__(self, session_params):
        self.test_source = "march_mad"
        self.test_target_source = "spreads"
        super(MMSessionCoverMarchMad, self).__init__(session_params)
        self.target_col = "Cover"
        self.target_source = "spreads"
        self.target_type = "multiclass"  # Tells if y is  'multiclass' or 'continuous'
        self.target_op = mm.MMTeamTransform.cover  # MMTeamTransform to calculate the target col
        self.report_cols = ['Season',  # Col names to add to predict set for post reporting (other than target_col)
                            'DayNum',
                            'Segment',
                            'T1_TeamID',
                            'T2_TeamID',
                            self.target_col]


class MMSessionCoverRegularSeason(MMSessionBase):
    def __init__(self, session_params):

        # TODO: Why am I calling the super after setting test/test_target sources?
        self.test_source = "cur_season"
        self.test_target_source = "spreads"
        super(MMSessionCoverRegularSeason, self).__init__(session_params)
        self.target_col = "Cover"
        self.target_source = "spreads"
        self.target_type = "multiclass"  # Tells if y is  'multiclass' or 'continuous'
        self.target_op = mm.MMTeamTransform.cover  # MMTeamTransform to calculate the target col
        self.report_cols = ['Season',  # Col names to add to predict set for post reporting (other than target_col)
                            'DayNum',
                            'Segment',
                            'T1_TeamID',
                            'T2_TeamID',
                            self.target_col]
        self.extend_report_cols()

        self.preprocessor = Pipeline([("scaler", StandardScaler()),
                                      ('poly', PolynomialFeatures(degree=2, include_bias=False))])

        self.njobs = 1  # << for debugging



