# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import glob
import logging
import os
import pathlib
import types
import warnings
import json

import google.cloud.storage as gcs
import h2o
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm.autonotebook import tqdm

pd.options.mode.chained_assignment = None  # default='warn'


class Util:

    # ------------------------------------------------------------------------
    # Opens the source for the feature_def calc
    # ------------------------------------------------------------------------
    @classmethod
    def get_source_data(cls, source, path='', local_root='./', gcs_root='/ops'):

        source_data = None

        if isinstance(source, pd.DataFrame):
            source_data = source

        elif source == 'reg_season':
            source_data = cls._get_source_df(file_name='MRegularSeasonCompactResults.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'cur_season':
            source_data = cls._get_source_df(file_name='MRegularSeasonCompactResults_current.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'reg_detail':
            source_data = cls._get_source_df(file_name='MRegularSeasonDetailedResults.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'cur_detail':
            source_data = cls._get_source_df(file_name='MRegularSeasonDetailedResults_current.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'march_mad':
            source_data = cls._get_source_df(file_name='MNCAATourneyCompactResults.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'seeds':
            source_data = cls._get_source_df(file_name='MNCAATourneySeeds.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'ordinals':
            source_data = cls._get_source_df(file_name='MMasseyOrdinals.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'kaggle_sub':
            source_data = cls._get_source_df(file_name='MSampleSubmissionStage1_2022_Spread.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'future':
            source_data = cls._get_source_df(file_name='future.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'spreads':
            source_data = cls._get_source_df(file_name='MSpreads.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'pMOV_b3':
            source_data = cls._get_pred_feature_source(feature_name=source)

        elif source == 'pMOV_b5':
            source_data = cls._get_pred_feature_source(feature_name=source)

        elif source == 'pVSpread':
            source_data = cls._get_pred_feature_source(feature_name=source)

        elif source == 'pMOV_b5_abs':
            source_data = cls._get_pred_feature_source(feature_name=source)

        elif source == 'pVSpread_abs':
            source_data = cls._get_pred_feature_source(feature_name=source)

        elif source == 'pCover_B5Abs':
            source_data = cls._get_pred_feature_source(feature_name=source)

        elif source == 'team_spellings':
            source_data = cls._get_source_df(file_name='MTeamSpellings.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'player_stats':
            source_data = cls._get_source_df(file_name='MPlayerStats.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'absences':
            source_data = cls._get_source_df(file_name='MAbsences.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'model_db':
            source_data = cls._get_source_df(file_name='model_database.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'feature_db':
            source_data = cls._get_source_df(file_name='feature_database.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'rosters':
            source_data = cls._get_source_json(file_name='MRosters.json', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'seasons':
            source_data = cls._get_source_df(file_name='MSeasons.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'mapi_teams':
            source_data = cls._get_source_df(file_name='MAPITeams.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        elif source == 'teams':
            source_data = cls._get_source_df(file_name='MTeams.csv', path=path, local_root=local_root, gcs_root=gcs_root)

        else:
            logging.error('Invalid source!')

        return source_data

    @classmethod
    def _get_source_df(cls, file_name=None, path='', local_root='',  gcs_root='ops/', bucket_name='mm21'):

        file_name = os.path.join(path, file_name)
        local_name = os.path.join(local_root, file_name)
        gcs_name = os.path.join(gcs_root, file_name)

        if not os.path.exists(local_name):
            try:
                cls.download(bucket_name=bucket_name, gcs_name=gcs_name, local_name=local_name)

            except Exception:
                logging.error(f'Cannot find source: {file_name} locally or on google cloud storage!')

        df = pd.read_csv(local_name, encoding='ISO-8859-1')

        return df

    @classmethod
    def _get_source_json(cls, file_name=None, path=None, local_root='', gcs_root='ops/', bucket_name='mm21'):

        file_name = os.path.join(path, file_name)
        local_name = os.path.join(local_root, file_name)
        gcs_name = os.path.join(gcs_root, file_name)

        if not os.path.exists(local_name):
            try:
                cls.download(bucket_name=bucket_name, gcs_name=gcs_name, local_name=local_name)

            except Exception:
                logging.error(f'Cannot find source: {file_name} locally or on google cloud storage!')

        with open(local_name, "r") as infile:
            data = json.load(infile)

        return data

    @classmethod
    def _get_pred_feature_source(cls, feature_name=None, local_root='', gcs_root='ops/'):

        feature_db = cls.get_source_data(source='feature_db', path='features', local_root=local_root, gcs_root=gcs_root)
        feature_db = feature_db.loc[feature_db["FeatureName"] == feature_name, :]

        df_source = pd.DataFrame()

        for feature_file, season, segment in zip(feature_db['FilePath'], feature_db['Season'], feature_db['Segment']):
            df_tmp = pd.read_csv(f'features/{feature_file}', encoding='ISO-8859-1')
            df_tmp['Season'] = season
            df_tmp['Segment'] = segment
            if not df_tmp.empty:
                df_source = pd.concat([df_source, df_tmp])

        return df_source

    @classmethod
    def delete_saved_artifact(cls, path, sub_path='features'):
        if sub_path in ['features', 'models', 'preds', 'submit']:
            if sub_path in ['preds', 'submit']:
                sub_path = os.path.join('../output', sub_path)
            delete_path = os.path.join(path, sub_path)

            items = 0
            for root, dirs, files in os.walk(os.path.join(delete_path), topdown=False):
                for name in files:
                    if name.endswith('.csv'):
                        os.remove(os.path.join(root, name))
                        items += 1
                for name in dirs:
                    os.rmdir(os.path.join(root, name))

            logging.info(f'Removed {items} files from {sub_path} and all sub folders...')
        else:
            logging.warning(
                f"Attempted deletion of {sub_path} not permitted, use one of: ['features', 'models', 'preds', 'submit']")

    @classmethod
    def delete_saved_feature(cls, path, feature, segment_count=None, lag=None, window=None):
        dtypes = {
            "FeatureName": str,
            "SegmentCount": "Int64",
            "Season": "Int64",
            "Segment": "Int64",
            "Lag": "Int64",
            "Window": "Int64",
            "FilePath": str
        }

        db_path = os.path.join(path, "feature_database.csv")
        df_feature_db = pd.read_csv(db_path, dtype=dtypes)

        if lag is None and window is None:
            if segment_count is None:
                df_mask = df_feature_db["FeatureName"] == feature

            else:
                df_mask = (df_feature_db["FeatureName"] == feature) & \
                          (df_feature_db["SegmentCount"] == segment_count)

        elif lag is None:
            if segment_count is None:
                df_mask = (df_feature_db["FeatureName"] == feature) & \
                          (df_feature_db["Window"] == window)
            else:
                df_mask = (df_feature_db["FeatureName"] == feature) & \
                          (df_feature_db["SegmentCount"] == segment_count) & \
                          (df_feature_db["Window"] == window)

        else:
            if segment_count is None:
                df_mask = (df_feature_db["FeatureName"] == feature) & \
                          (df_feature_db["Lag"] == lag) & \
                          (df_feature_db["Window"] == window)
            else:
                df_mask = (df_feature_db["FeatureName"] == feature) & \
                          (df_feature_db["SegmentCount"] == segment_count) & \
                          (df_feature_db["Lag"] == lag) & \
                          (df_feature_db["Window"] == window)

        df_removal = df_feature_db[df_mask]
        df_feature_db = df_feature_db[~df_mask]

        if df_removal.empty:
            if lag is None and window is None:
                logging.info(f"No database entries of {feature} found, moving on...")

            elif lag is None:
                logging.info(f"No database entries of {feature} | Window: {window} found, moving on...")

            else:
                logging.info(f"No database entries of {feature} | Lag: {lag} | Window: {window} found, moving on...")

        else:
            for i, row in df_removal.iterrows():
                try:
                    os.remove(os.path.join(path, row["FilePath"]))
                except FileNotFoundError:
                    continue

            df_feature_db.to_csv(db_path, index=False)

            if lag is None and window is None:
                logging.info(f"Removed all files and database entries of {feature}")

            elif lag is None:
                logging.info(f"Removed all files and database entries of {feature} | Window: {window}")

            else:
                logging.info(f"Removed all files and database entries of {feature} | Lag: {lag} | Window: {window}")

    # @classmethod
    # def delete_all_saved_artifacts(cls, path):
    #
    #     items = 0
    #
    #     for sub_path in ['features', 'models', 'preds', 'submit']:
    #         if sub_path in ['preds', 'submit']:
    #             sub_path = os.path.join('../output', sub_path)
    #         delete_path = os.path.join(path, sub_path)
    #
    #         for root, dirs, files in os.walk(os.path.join(delete_path), topdown=False):
    #             for name in files:
    #                 if name.endswith('.csv') or sub_path == 'models':
    #                     os.remove(os.path.join(root, name))
    #                     items += 1
    #             for name in dirs:
    #                 os.rmdir(os.path.join(root, name))
    #
    #         logging.info(f'Removed {items} files from {sub_path} and all sub folders...')

    @classmethod
    def swap_extend(cls, df):
        df_swap = df.copy()
        df_swap.loc[df['WLoc'] == 'H', 'WLoc'] = 'A'
        df_swap.loc[df['WLoc'] == 'A', 'WLoc'] = 'H'
        df = df.rename(columns={"WLoc": "location"})
        df_swap = df_swap.rename(columns={"WLoc": 'location'})
        df = df.rename(columns=lambda x: x.replace("W", "T1_").replace("L", "T2_"))
        df_swap = df_swap.rename(columns=lambda x: x.replace("W", "T2_").replace("L", "T1_"))
        df = pd.concat([df, df_swap]).sort_index().reset_index(drop=True)

        return df

    # doubles the data size by generalizing the W/L labels and reversing and replicating the data
    @classmethod
    def swap_extend_current(cls, df):
        df = df.copy()
        df_swap = df.rename(columns=lambda x: x.replace("T1_", "TBD"))
        df_swap = df_swap.rename(columns=lambda x: x.replace("T2_", "T1_"))
        df_swap = df_swap.rename(columns=lambda x: x.replace("TBD", "T2_"))
        df = pd.concat([df, df_swap]).sort_index().reset_index(drop=True)

        return df

    @classmethod
    def swap_extend_spreads(cls, df):
        df_swap = df.copy()
        df = df.rename(columns=lambda x: x.replace("W", "T1_").replace("L", "T2_"))
        df_swap = df_swap.rename(columns=lambda x: x.replace("W", "T2_").replace("L", "T1_"))
        df_swap["line"] = -df_swap["line"]
        df = pd.concat([df, df_swap])
        df = df.sort_index()
        df = df.reset_index(drop=True)

        return df

    @classmethod
    def swap_extend_preds(cls, df):
        exclude_cols = ['Season', 'DayNum', 'T1_TeamID', 'T2_TeamID']
        cols = [c for c in df.columns.values.tolist() if c not in exclude_cols]

        df_swap = df.copy()
        df_swap = df_swap.rename(columns={"T1_TeamID": "tmp", "T2_TeamID": "T1_TeamID"})
        df_swap = df_swap.rename(columns={"tmp": "T2_TeamID"})
        df_swap[cols] = -df_swap[cols]
        df_swap = df_swap.sort_index().reset_index(drop=True)
        df = pd.concat([df, df_swap]).sort_index().reset_index(drop=True)

        return df

    # --------------------------------------------------------------
    #  Chunks seasons into segments
    # -----------------------------------------------------------
    @classmethod
    def chunk_seasons(cls, df, segment_count=4):

        df['Segment'] = -1

        for season in range(df['Season'].min(), df['Season'].max() + 1):

            chunk_size = len(df[df['Season'] == season]) // segment_count

            first_day = df.loc[df['Season'] == season, 'DayNum'].min()
            first_index = df[df['Season'] == season].index.min()
            for i in range(segment_count):
                if i == segment_count - 1:
                    last_day = df.loc[df['Season'] == season, 'DayNum'].max()
                else:
                    last_day = df['DayNum'][first_index + (i + 1) * chunk_size]

                df.loc[((df['Season'] == season) & (df['DayNum'] >= first_day) & (
                            df['DayNum'] <= last_day)), 'Segment'] = i
                first_day = last_day + 1

        return df

    # --------------------------------------------------------------
    #  Crops a data set on Season and DayNum
    # --------------------------------------------------------------
    @classmethod
    def crop(cls, df, start=None, end=None, segment_count=4):

        if len(start) == 2 and len(end) == 2:
            if start == end:
                df = df[((df['Season'] == start[0]) & (df['Segment'] == start[1]))]
            else:
                df = df[((df['Season'] == start[0]) & (df['Segment'] >= start[1]) & (df['Segment'] <= segment_count)) |
                        ((df['Season'] > start[0]) & (df['Season'] < end[0])) |
                        ((df['Season'] == end[0]) & (df['Segment'] <= end[1]))]

        elif len(start) == 3 and len(end) == 3:
            df = df[((df['Season'] == start[0]) & (df['DayNum'] >= start[2]) & (df['DayNum'] <= end[2])) |
                    ((df['Season'] > start[0]) & (df['Season'] < end[0])) |
                    ((df['Season'] == end[0]) & (df['DayNum'] <= end[2]) & (df['DayNum'] >= start[2]))]

        else:
            raise ValueError("start and end must be tuples of lengths 2 or 3")

        df = df.reset_index(drop=True)

        return df

    @classmethod
    def split(cls, df, start=None, end=None, start_cut=None, end_cut=None, segment_count=4):

        data_type_list = df.dtypes

        if start_cut == end_cut:
            df_cut = df[((df['Season'] == start_cut[0]) & (df['Segment'] == start_cut[1]))]
        else:
            df_cut = df[((df['Season'] == start_cut[0]) & (df['Segment'] >= start_cut[1]) & (df['Segment'] <= segment_count)) |
                        ((df['Season'] > start_cut[0]) & (df['Season'] < end_cut[0])) |
                        ((df['Season'] == end_cut[0]) & (df['Segment'] <= end_cut[1]))]

        if start == end:
            df = df[((df['Season'] == start[0]) & (df['Segment'] == start[1]))]
        else:
            df = df[((df['Season'] == start[0]) & (df['Segment'] >= start[1]) & (df['Segment'] <= segment_count)) |
                    ((df['Season'] > start[0]) & (df['Season'] < end[0])) |
                    ((df['Season'] == end[0]) & (df['Segment'] <= end[1]))]

        df = df[~df.isin(df_cut)].dropna()

        for column, data_type in data_type_list.items():
            df[column] = df[column].astype(data_type)
            df_cut[column] = df_cut[column].astype(data_type)

        df = df.reset_index(drop=True)
        df_cut = df_cut.reset_index(drop=True)

        return df, df_cut

    @classmethod
    def adjust_for_lag_and_window(cls, start=None, end=None, lag=1, window=4, segment_count=4):

        if lag < 0 | window < 0:
            logging.error('Error: lags and windows must be positive!')

        # adjust start for window and lag, and end for lag
        end = cls.adjust_date(end, adj=-lag, segment_count=segment_count)
        start = cls.adjust_date(start, adj=-lag - window + 1, segment_count=segment_count)

        return start, end

    @classmethod
    def adjust_date(cls, d, adj=1, segment_count=4):

        year = d[0]
        seg = d[1]
        delta = seg + adj
        seg = delta % segment_count
        year = year + (delta // segment_count)
        adj_date = (year, seg)

        return adj_date

    @classmethod
    def get_window(cls, start=None, window=4, segment_count=4):

        delta = start[1] + window
        end_seg = delta % segment_count
        end_year = start[0] + delta // segment_count

        if len(start) == 2:
            end = (end_year, end_seg)

        elif len(start) == 3:
            end = (end_year, end_seg, start[2])

        else:
            raise ValueError("start and end must be tuples of lengths 2 or 3")

        return start, end

    @classmethod
    def calc_profit(cls, df_results, threshold):
        df_results["Profit"] = 0
        df_results.loc[(df_results["Cover"] == 1) & (df_results["Mean"] >= threshold), "Profit"] = 1
        df_results.loc[(df_results["Cover"] == 2) & (df_results["Mean"] >= threshold), "Profit"] = -1.1

        return df_results["Profit"].sum()

    @classmethod
    def calc_ev(cls, df_results, threshold):
        df_results["EV"] = 0
        df_results.loc[df_results["Mean"] >= threshold,
                       "EV"] = df_results.loc[df_results["Mean"] >= threshold, "Mean"] - \
                               1.1 * (1 - df_results.loc[df_results["Mean"] >= threshold, "Mean"])

        return df_results["EV"].sum()

    @classmethod
    def calc_std(cls, df_results, threshold):
        df_results["Variance"] = 0
        df_results.loc[df_results["Mean"] >= threshold,
                       "Variance"] = 1.1 * df_results.loc[df_results["Mean"] >= threshold, "Mean"] * (
                1 - df_results.loc[df_results["Mean"] >= threshold, "Mean"])

        return df_results["Variance"].sum() ** 0.5

    @classmethod
    def calc_nbets(cls, df_results, threshold):
        return df_results.loc[df_results["Mean"] >= threshold, "Mean"].count()

    @classmethod
    def calc_win_pct(cls, df_results, threshold):
        return df_results.loc[(df_results["Cover"] == 1) & (df_results["Mean"] >= threshold), "Mean"].count() / \
               df_results.loc[df_results["Mean"] >= threshold, "Mean"].count()

    @classmethod
    def unswap_results(cls, df, report_cols, include_ids=False):
        base_cols = ["Season", "Segment", "DayNum", "T1_TeamName", "T2_TeamName", "p1", "Cover"]

        if include_ids:
            feature_cols = list(set(report_cols).difference(base_cols + ["p2"]))
        else:
            feature_cols = list(set(report_cols).difference(base_cols + ["T1_TeamID", "T2_TeamID", "p2"]))

        feature_cols.sort()
        df_results = pd.DataFrame(columns=["Season", "Segment", "DayNum", "Team Of Interest",
                                           "Opponent", "T1.p1", "Cover"] + feature_cols)
        df_results[["Season", "Segment", "DayNum", "Team Of Interest", "Opponent", "T1.p1", "Cover"] + feature_cols] = df[base_cols + feature_cols]

        df_results = df_results.merge(df[["Season", "Segment", "DayNum", "T1_TeamName", "T2_TeamName", "p2"]],
                                      left_on=["Season", "DayNum", "Opponent", "Team Of Interest"],
                                      right_on=["Season", "DayNum", "T1_TeamName", "T2_TeamName"])
        df_results = df_results.drop(columns=["T1_TeamName", "T2_TeamName", "Segment_y"])
        df_results = df_results.rename(columns={"p2": "T2.p2", "Segment_x": "Segment"})
        df_results["Mean"] = (df_results["T1.p1"] + df_results["T2.p2"]) / 2
        df_results = df_results[["Season", "Segment", "DayNum", "Team Of Interest", "Opponent", "Mean", "Cover"] + feature_cols]
        df_results = df_results.reset_index(drop=True)

        return df_results

    @classmethod
    def analyze_profits(cls, df, report_cols, thresholds=np.arange(0.5, 0.91, 0.01)):
        warnings.filterwarnings("ignore", message="invalid value encountered in long_scalars", category=RuntimeWarning)

        df_results = cls.unswap_results(df, report_cols)

        df_profit = pd.DataFrame(columns=["Threshold"])

        df_profit["Threshold"] = thresholds
        df_profit["Profit"] = df_profit.apply(lambda x: cls.calc_profit(df_results, x["Threshold"]), axis=1)
        df_profit["Expected Value"] = df_profit.apply(lambda x: cls.calc_ev(df_results, x["Threshold"]), axis=1)
        df_profit["Standard Deviation (EV)"] = df_profit.apply(lambda x: cls.calc_std(df_results, x["Threshold"]), axis=1)
        df_profit["# of Bets"] = df_profit.apply(lambda x: cls.calc_nbets(df_results, x["Threshold"]), axis=1)
        df_profit["Units Risked"] = df_profit["# of Bets"] * 1.1
        df_profit["%ROI"] = df_profit["Profit"] / df_profit["Units Risked"]
        df_profit["Win %"] = df_profit.apply(lambda x: cls.calc_win_pct(df_results, x["Threshold"]), axis=1)
        df_profit["MOE (Win %)"] = 1.96 * ((df_profit["Win %"] * (1 - df_profit["Win %"])) / df_profit["# of Bets"]) ** 0.5
        # df_profit = df_profit[df_profit["# of Bets"] != 0]
        # df_profit["True %ROI"] = df_profit["Profit"] / df_profit["Units Risked"]
        # df_profit["Expected %ROI"] = df_profit["Expected Value"] / df_profit["Units Risked"]

        df_profit = df_profit.fillna(0)

        warnings.resetwarnings()

        return df_results, df_profit

    @classmethod
    def graph_profit_risk(cls, df, **kwargs):
        df_results, df_profit = cls.analyze_profits(df, **kwargs)

        fig, ax = plt.subplots()

        ax.set_ylabel("Profit (Units)")
        ax.set_xlabel("Standard Deviation (Units)")
        ax.set_title("Profit and % ROI vs. Risk")
        ax.plot(df_profit["Standard Deviation (EV)"], df_profit["Profit"], label="Profit")

        return df_results, df_profit, fig

    # experimental: builds a models_db by locking in the model folder and the prediction folder to infer what the
    # model_database should have in it.  to be used with fix_model_db.py to rebuild the model_database if it is lost
    @classmethod
    def build_model_db(cls, model_path, pred_path):

        df = pd.DataFrame(columns=['Season', 'SegmentCount', 'Segment', 'Model_File'])

        model_files = glob.glob(model_path + '/**/*/Step_000/*', recursive=True)

        for model_file in tqdm(model_files, desc='Scanning model files: '):

            run_id = pathlib.Path(model_file).parts[1]

            if os.path.exists(f'{pred_path}/{run_id}'):
                pred_files = glob.glob(f'{pred_path}/**/{run_id}/run_*.csv', recursive=True)
                for pred_file in pred_files:
                    season = int(pred_file[-12:].split('.')[0][:4])
                    segment = int(pred_file[-12:].split('.')[0][5:6])
                    segment_count = int(pred_file[-12:].split('.')[0][7:8])
                    db_entry = {'Season': season,
                                'SegmentCount': segment_count,
                                'Segment': segment,
                                'Model_File': model_file}
                    df = df.append(db_entry, ignore_index=True)

        return df.sort_values(by=['Season', 'Segment']).reset_index(drop=True)

    @classmethod
    def convert_models_to_mojo(cls, model_path):
        model_files = glob.glob(model_path + '/**/*/Step_000/*', recursive=True)

        for file in tqdm(model_files):
            if file[-3:] != "zip":
                try:
                    bin_model = h2o.load_model(file)
                    path = pathlib.Path(file)
                    bin_model.save_mojo(path=path.parent.name, force=True, filename=f"{path.name}.zip")
                except h2o.exceptions.H2OResponseError:
                    logging.warning(f"H2O version mismatch on {file}")

    @classmethod
    def update_feature_db(cls, season=2004, segment=0, feature_name='', segment_count=4, lag=None, window=None,
                          db_path='', feature_path='', local_root='./', gcs_root='ops/'):

        local_path = os.path.join(local_root, db_path)
        gcs_path = os.path.join(gcs_root, db_path)

        dtypes = {
            "FeatureName": str,
            "SegmentCount": "Int64",
            "Season": "Int64",
            "Segment": "Int64",
            "Lag": "Int64",
            "Window": "Int64",
            "FilePath": str
        }

        if os.path.exists(local_path):
            df = Util.get_source_data('feature_db', path=db_path, local_root=local_root, gcs_root=gcs_root)
            # df = pd.read_csv(db_path, dtype=dtypes)
            if lag is None or window is None:
                old_row = df.loc[(df["Season"] == season) &
                                 (df["Segment"] == segment) &
                                 (df["FeatureName"] == feature_name) &
                                 (df["SegmentCount"] == segment_count), :]

                if not old_row.empty:
                    df = df.drop(df[(df["Season"] == season) &
                                    (df["Segment"] == segment) &
                                    (df["FeatureName"] == feature_name) &
                                    (df["SegmentCount"] == segment_count)].index)

            else:
                old_row = df.loc[(df["Season"] == season) &
                                 (df["Segment"] == segment) &
                                 (df["FeatureName"] == feature_name) &
                                 (df["SegmentCount"] == segment_count) &
                                 (df["Lag"] == lag) &
                                 (df["Window"] == window), :]

                if not old_row.empty:
                    df = df.drop(df[(df["Season"] == season) &
                                    (df["Segment"] == segment) &
                                    (df["FeatureName"] == feature_name) &
                                    (df["SegmentCount"] == segment_count) &
                                    (df["Lag"] == lag) &
                                    (df["Window"] == window)].index)

            row = pd.DataFrame(
                {'Season': season, 'Segment': segment, 'FeatureName': feature_name, 'SegmentCount': segment_count,
                 'Lag': lag, 'Window': window, 'FilePath': [feature_path]})

            df = pd.concat([df, row], ignore_index=True)
            df = df.sort_values(by=["FeatureName", "SegmentCount", "Season", "Segment"], ignore_index=True)
            df.to_csv(os.path.join(local_path, 'feature_database.csv'), index=False)
            logging.info(f'Season: {season} | Segment: {segment} | Feature: {feature_name} >> saved to feature database..')
        else:
            logging.error(f"No feature_database.csv found at {db_path}. Can't bootstrap feature_database.csv with this utility...")
            feature_path = None

        return feature_path

    @classmethod
    def _download_from_csv_list(cls, source=None, path=None, local_root='./', gcs_root='ops/'):

        downloaded = True

        # here we support passing a list or certain dfs
        if isinstance(source, list):
            pbar = tqdm(os.path.join(path, x) for x in source)

        elif isinstance(source, types.GeneratorType):
            source_list = []
            for root, folders, files in source:
                for file in files:
                    source_list.append(os.path.join(root[2:], file))
            pbar = tqdm(source_list, desc=f"Uploading files from {path}: ")

        else:
            df = cls.get_source_data(source=source, path=path, local_root=local_root, gcs_root=gcs_root)
            if source == 'feature_db':
                df['FilePath'] = df['FilePath'].apply(lambda x: os.path.join(f'{path}', x))
                pbar = tqdm(df['FilePath'].tolist(), desc="Downloading features: ")
            elif source == 'model_db':
                pbar = tqdm(df['Model_File'].tolist(), desc="Downloading models: ")
            else:
                logging.error(f'Invalid csv source: {source}!!')

        for file_name in pbar:

            if not os.path.exists(os.path.join(local_root, file_name)):
                if not os.path.exists(os.path.dirname(os.path.join(local_root, file_name))):
                    os.makedirs(os.path.dirname(os.path.join(local_root, file_name)))
                ok = cls.download(gcs_name=os.path.join(gcs_root, file_name), local_name=os.path.join(local_root, file_name))
                if not ok:
                    logging.error(f'Failed downloading: {file_name}!!')
                    downloaded = False
                    break

        return downloaded

    @classmethod
    def _upload_from_path_list(cls, source=None, path=None, local_root='./', gcs_root='ops/', no_pbar=False):

        uploaded = True

        # here we support passing a list or certain dfs
        if isinstance(source, list):
            source = [os.path.join(path, x) for x in source]
            pbar = tqdm(source, desc=f"Uploading files from {path}: ", disable=no_pbar)

        elif isinstance(source, types.GeneratorType):
            source_list = []
            for root, folders, files in source:
                for file in files:
                    source_list.append(os.path.join(root[2:], file))
            pbar = tqdm(source_list, desc=f"Uploading files from {path}: ", disable=no_pbar)

        else:
            df = cls.get_source_data(source=source, path=path, local_root=local_root, gcs_root=gcs_root)

            if source == 'feature_db':
                df['FilePath'] = df['FilePath'].apply(lambda x: os.path.join(path, x))
                pbar = tqdm(df['FilePath'].tolist(), desc="Uploading features: ")
            elif source == 'model_db':
                pbar = tqdm(df['Model_File'].tolist(), desc="Uploading models: ")
            else:
                logging.error(f'Invalid source: {source}!!')

        for file_name in pbar:
            if os.path.isdir(file_name):
                ok = cls._upload_from_path_list(source=os.listdir(file_name), path=file_name, no_pbar=True)
            else:
                ok = cls.upload(gcs_name=os.path.join(gcs_root, file_name), local_name=os.path.join(local_root, file_name))

            if not ok:
                uploaded = False
                logging.error(f'Failed uploading: {file_name}!!')
                break

        return uploaded

    @classmethod
    def prepare_environment(cls, local_root='./', gcs_root='/ops'):

        # make sure we have good paths to load data
        Util.make_paths(path_list=[f'{local_root}data/our_data', f'{local_root}output/preds', f'{local_root}models', f'{local_root}features/data'])

        back_up = cls.get_source_data('model_db', path='models/', local_root=local_root, gcs_root=gcs_root)
        back_up.to_csv(f'{local_root}models/last.model_database.csv', index=False)
        downloaded = cls._download_from_csv_list(source='feature_db', path='features/', local_root=local_root, gcs_root=gcs_root)
        if not downloaded:
            logging.error('Failed downloading features...')
        else:
            back_up = cls.get_source_data('feature_db')
            back_up.to_csv(f'{local_root}models/last.feature_database.csv', index=False)
            logging.info('Environment has been prepared on the target machine...')

    @classmethod
    def refresh_model_db(cls, local_root='./', gcs_root='ops/'):
        gcs_loaded = cls.download(gcs_name=f"{gcs_root}models/model_database.csv", local_name=f"{local_root}models/gcs_model_database.csv")
        if gcs_loaded:
            df_gcs_db = pd.read_csv(f"{local_root}models/gcs_model_database.csv")
            # path=f"{local_root}models/"
            df_local_db = cls.get_source_data("model_db", path='models/', local_root=local_root, gcs_root=gcs_root)

            df_new_db = pd.concat([df_gcs_db, df_local_db])
            df_new_db = df_new_db.sort_values(by="Score")
            df_new_db = df_new_db.drop_duplicates(subset=["Season", "Segment", "Segment_Count", "Target", "Description",
                                                          "Metric"], keep="first")
            df_new_db = df_new_db.sort_values(by=["Target", "Description", "Segment_Count", "Season", "Segment"],
                                              ignore_index=True)

            df_new_db.to_csv(f"{local_root}models/model_database.csv", index=False)
            os.remove(f"{local_root}models/gcs_model_database.csv")
            return True

        else:
            logging.error("Unable to find gcs copy of model_database.csv")
            return False

    @classmethod
    def refresh_feature_db(cls, local_root='./', gcs_root='ops/'):
        gcs_loaded = cls.download(gcs_name=f"{gcs_root}features/feature_database.csv",
                                  local_name=f"{local_root}features/gcs_feature_database.csv")
        if gcs_loaded:
            df_gcs_db = pd.read_csv(f"{local_root}features/gcs_feature_database.csv")
            df_local_db = cls.get_source_data("feature_db", path=f"features/", local_root=local_root, gcs_root=gcs_root)

            df_new_db = pd.concat([df_gcs_db, df_local_db])
            df_new_db = df_new_db.drop_duplicates()
            df_new_db = df_new_db.sort_values(by=["FeatureName", "SegmentCount", "Season", "Segment"],
                                              ignore_index=True)

            df_new_db.to_csv(f"{local_root}features/feature_database.csv", index=False)
            os.remove(f"{local_root}features/gcs_feature_database.csv")
            return True

        else:
            logging.error("Unable to find gcs copy of feature_database.csv")
            return False

    @classmethod
    def save_environment(cls, local_root='', gcs_root='ops/'):

        ok_to_delete = True
        if cls.refresh_model_db(local_root=local_root, gcs_root=gcs_root):
            if not cls.upload(local_name=f'{local_root}models/model_database.csv', gcs_name=f'{gcs_root}models/model_database.csv', replace=True):
                ok_to_delete = False
            if not cls.upload(local_name=f'{local_root}models/last.model_database.csv', gcs_name=f'{gcs_root}models/last.model_database.csv', replace=True):
                ok_to_delete = False
            if not cls._upload_from_path_list(source='model_db', path=f'{local_root}models/'):
                ok_to_delete = False
        else:
            ok_to_delete = False

        if cls.refresh_feature_db(local_root=local_root, gcs_root=gcs_root):
            if not cls.upload(local_name=f'{local_root}features/feature_database.csv', gcs_name=f'{gcs_root}features/feature_database.csv', replace=True):
                ok_to_delete = False
            if not cls.upload(local_name=f'{local_root}features/last.feature_database.csv', gcs_name=f'{gcs_root}features/last.feature_database.csv', replace=True):
                ok_to_delete = False
            if not cls._upload_from_path_list(source='feature_db', path=f'{local_root}features/'):
                ok_to_delete = False
        else:
            ok_to_delete = False

        if not cls._upload_from_path_list(source=os.listdir(f'{local_root}data'), path='data'):
            ok_to_delete = False

        if not cls._upload_from_path_list(source=os.walk(f'{local_root}output'), path='output/'):
            ok_to_delete = False


        if ok_to_delete:
            logging.info('Environment has been saved to google cloud storage...')
        else:
            logging.error('At least one upload failed...')

        return ok_to_delete

    @classmethod
    def clean_up_environment(cls):

        # if saved:
        #     # delete local files
        #     for f in os.listdir('./models'):
        #         path = os.path.join('./models', f)
        #         if os.path.isdir(path):
        #             shutil.rmtree(path)
        #         else:
        #             try:
        #                 cls.upload(path, )
        #             os.remove(path)
        pass
        # delete local files
        # for f in os.listdir('./models'):
        #     path = os.path.join('./models', f)
        #     if os.path.isdir(path):
        #         shutil.rmtree(path)
        #     else:
        #         try:
        #             cls.upload(path, )
        #         os.remove(path)

        # if saved:
        #     # delete local files
        #     for f in os.listdir('./features'):
        #         path = os.path.join('./features', f)
        #         if os.path.isdir(path):
        #             pass # need to leave the ./feature/data folder in place
        #         else:
        #             os.remove(os.path.join('./features', f))


    @classmethod
    def download(cls, bucket_name='mm21', gcs_name='', local_name=''):

        successful = True

        try:
            storage_client = gcs.Client()
            bucket = storage_client.bucket(bucket_name)
            blob = bucket.blob(gcs_name)

            blob.download_to_filename(local_name)
            logging.debug(f'Downloaded {gcs_name} to {local_name}...')

        except Exception:
            successful = False
            logging.error(f'Error downloading gcs file: {gcs_name} to {local_name}!')

        return successful


    @classmethod
    def upload(cls, bucket_name='mm21', local_name='', gcs_name='', replace=False):

        successful = True
        try:
            storage_client = gcs.Client()
            bucket = storage_client.bucket(bucket_name)
            blob = bucket.blob(gcs_name)

            if replace:
                blob.upload_from_filename(local_name)
                logging.info(f'Uploaded {local_name} to gs://{bucket_name}/{gcs_name}...')
            if not blob.exists():
                blob.upload_from_filename(local_name)
                logging.info(f'Uploaded {local_name} to gs://{bucket_name}/{gcs_name}...')
        except Exception as e:
            successful = False
            logging.error(f'Error uploading local file: {local_name} to gs://{bucket_name}/{gcs_name}!')

        return successful

    @classmethod
    def gcs_file_exists(cls, bucket_name='mm21', gcs_name=''):
        storage_client = gcs.Client()
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(gcs_name)
        return blob.exists()

    @classmethod
    def make_paths(cls, path_list=None):
        for path in path_list:
            if not os.path.exists(path):
                os.makedirs(path)

