# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
import logging
import os
import pathlib
import time
import uuid

import h2o
import pandas as pd
from h2o.automl import H2OAutoML

import src.mmutil as mu
from src.mmlookup import MMLookup


# -----------------------------------------------------------------------------
#
#   Name:      MMTrain
#
#   Purpose:   Primary train object. Runs training, prediction.
#
#
#   Usage:     config = MMSessionConfig(...) <-- see source code for args
#              trainer = MMTrain(config)
#              aml, test, val = trainer.train()
#
# -----------------------------------------------------------------------------
class MMTrain:

    def __init__(self, config):
        self.config = config  # see MMSessionConfig
        self.best_models = ''
        self.aml = ''
        self.config.run_id = self.set_run_id(self.config.run_id)

        # get model database
        model_db_path = os.path.join(self.config.local_root, self.config.model_path, 'model_database.csv')
        if os.path.exists(model_db_path):
            self.model_db = mu.Util.get_source_data('model_db', path=self.config.model_path, gcs_root=self.config.gcs_root, local_root=self.config.local_root)
        else:
            logging.debug(f'No model database found at: {model_db_path} ')
            self.model_db = pd.DataFrame(columns=["Season", "Segment", "Segment_Count", "Target", "Description",
                                                  "Metric", "Score", "Model_File"])

    # ------------------------------------------------------------------------
    # Runs a single train step
    # ------------------------------------------------------------------------
    def train(self, mojo=True):

        # fetch data sets
        train, test, preds = self.assemble_data()

        # get features
        features = self.set_train_cols(train)

        # get targets and conform to h2o expectations
        train = self.conform_targets(train)
        test = self.conform_targets(test)

        if self.config.future:
            preds = self.conform_targets(preds)

        # launch AML
        self.aml = H2OAutoML(nfolds=self.config.nfolds,
                             max_runtime_secs=self.config.max_runtime_secs,
                             max_runtime_secs_per_model=self.config.max_runtime_secs_per_model,
                             max_models=self.config.max_models,
                             stopping_metric=self.config.stopping_metric,
                             stopping_rounds=self.config.stopping_rounds,
                             seed=self.config.seed,
                             exclude_algos=self.config.exclude_algos,
                             include_algos=self.config.include_algos,
                             exploitation_ratio=self.config.exploitation_ratio,
                             keep_cross_validation_predictions=self.config.keep_cross_validation_predictions,
                             keep_cross_validation_fold_assignment=self.config.keep_cross_validation_fold_assignment,
                             sort_metric=self.config.sort_metric,
                             export_checkpoints_dir=self.config.export_checkpoints_dir,
                             verbosity=self.config.verbosity,
                             balance_classes=self.config.balance_classes
                             )

        # train on a data set
        if self.config.future:
            self.aml.train(x=features, y=self.config.target_col, training_frame=train)
        else:
            validation_frame = test if self.config.future is False else preds
            self.aml.train(x=features, y=self.config.target_col, training_frame=train,
                           validation_frame=validation_frame)

        # record step information
        self.record_step_info(self.aml)

        # save model
        self.save_models(mojo)

        return test, preds

    # ------------------------------------------------------------------------
    # Calls the data object to get train/test/val data
    # ------------------------------------------------------------------------
    def assemble_data(self):

        train = self.config.data.get_train_data()
        test = self.config.data.get_test_data()
        pred = self.config.data.get_pred_data() if self.config.future else None

        return train, test, pred

    # ------------------------------------------------------------------------
    # Gets the appropriate columns into the train/test/val data
    # ------------------------------------------------------------------------
    def set_train_cols(self, t):

        # if the column list is an exclusion list
        if self.config.exclude:
            features = t.columns
            for exclusion in self.config.train_cols:
                features.remove(exclusion)
        else:
            features = self.config.train_cols

        return features

    # ------------------------------------------------------------------------
    # Conforms targets, binary targets require the call to asfactor() for h2o
    # ------------------------------------------------------------------------
    def conform_targets(self, x):
        if self.config.target_type == 'multiclass':
            x[self.config.target_col] = x[self.config.target_col].asfactor()
        return x

    def load_model(self, model_file=None, run_id=None, season=None, segment=None, target_name=None, description=None):

        # load the model based on the run_id and model_file
        if model_file is not None and run_id is not None:
            model = self.open_model_file(run_id=run_id, model_file=model_file)

        # load the model for the season, segment, and target_name
        elif season is not None and segment is not None and target_name is not None:
            if description is None:
                model_file = self.model_db.loc[(self.model_db['Season'] == season) & (self.model_db['Segment'] == segment) &
                                               (self.model_db['Segment_Count'] == self.config.segment_count) &
                                               (self.model_db['Target'] == target_name)]['Model_File']
            else:
                model_file = self.model_db.loc[(self.model_db['Season'] == season) & (self.model_db['Segment'] == segment) &
                                               (self.model_db['Target'] == target_name) & (self.model_db['Description'] == description)]['Model_File']

            model_file = model_file.reset_index(drop=True)
            run_id = pathlib.Path(model_file[0]).parts[1]
            model_file = pathlib.Path(model_file[0]).parts[-1]
            model = self.open_model_file(run_id=run_id, model_file=model_file, gcs_root=self.config.gcs_root, local_root=self.config.local_root)

        # otherwise, load the current best model from lb_master
        else:
            model_id = self.config.lb_master[self.config.lb_master.step == f'Step_{self.config.step_num:03d}'].model_id.values[0]
            model = h2o.get_model(model_id)

        return model

    # ------------------------------------------------------------------------
    # Loads the best model relevant to the current step
    # ------------------------------------------------------------------------
    def load_best_model(self):
        # get the model_id
        model_id = \
            self.config.lb_master[self.config.lb_master.step == f'Step_{self.config.step_num:03d}'].model_id.values[0]

        return h2o.get_model(model_id)

    # ------------------------------------------------------------------------
    # Saves the models in an organized way for further use
    # ------------------------------------------------------------------------
    def save_models(self, mojo):

        save_path = os.path.join(self.config.model_path + self.config.run_id,
                                 self.add_step_name(self.config.run_id)[-8:])

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        save_list = self.config.lb_master[
            self.config.lb_master.step == f'Step_{self.config.step_num:03d}'].model_id.values

        save_list = save_list[:5]

        for m in save_list:
            model = h2o.get_model(m)
            if mojo:
                model_path = model.save_mojo(path=save_path, force=True, filename=f"{m}.zip")
            else:
                model_path = h2o.save_model(model=model, path=save_path, force=True,
                                            export_cross_validation_predictions=True)

            self.update_model_db(model=model, model_path=model_path)

            logging.info(f'Run ID: {self.config.run_id[-4:]} Saved model {m} to {save_path}...')

    # ------------------------------------------------------------------------
    # Logging and saving function for the train step
    # ------------------------------------------------------------------------
    def record_step_info(self, aml):

        # add the step number to the run id for storing models and projections with helpful names
        run_id = self.add_step_name(self.config.run_id)

        # get the leaderboard
        lb = h2o.automl.get_leaderboard(aml, extra_columns="ALL")

        msgs = self._get_report_msgs(aml)

        for msg in msgs:
            logging.info(msg)

        # save full leaderboard to master leaderboard
        df_lb = lb.as_data_frame()
        df_lb['run_id'] = run_id[:-9]
        df_lb['step'] = run_id[-8:]
        df_lb = df_lb.head(5)  # <<< keep top 5 only

        if isinstance(self.config.lb_master, pd.DataFrame):
            self.config.lb_master = self.config.lb_master.append(df_lb)
        else:
            self.config.lb_master = df_lb

        self.config.lb_master = self.config.lb_master.reset_index(drop=True)

    # ------------------------------------------------------------------------
    # Sets the run_id
    # ------------------------------------------------------------------------
    def set_run_id(self, run_id=None):

        if run_id == '':
            run_id = 'run_' + time.strftime("%y%m%d_%h%m%s") + '_' + str(uuid.uuid1())[:4]
        else:
            run_id = run_id

        self.config.run_id = run_id
        logging.debug(f'Run ID {run_id}')

        return run_id

    # -----------------------------------------------------------------------------------
    # Callers running more than one step must call this function to keep track of steps
    # -----------------------------------------------------------------------------------
    def update_step_num(self):
        self.config.save_aml_sess(self.aml)
        self.config.step_num += 1
        return self.config.step_num

    # -----------------------------------------------------------------------------------
    # Adds the step name to a run_id
    # -----------------------------------------------------------------------------------
    def add_step_name(self, run_id):
        return f'{run_id}_Step_{self.config.step_num:03d}'

    # -----------------------------------------------------------------------------------
    # Runs a single predict step and saves the results with reporting info
    # -----------------------------------------------------------------------------------
    def predict(self, test, season=None, segment=None, target_name=None, description=None, run_id=None,
                model_file=None, future=False):

        # load the model
        model = self.load_model(season=season, segment=segment, target_name=target_name, description=description,
                                run_id=run_id, model_file=model_file)

        # make the predictions
        preds = model.predict(test)

        # combine with other info for analyzing predictions
        df_preds = self._annotate_analysis_info(preds.as_data_frame(), future=future)

        # save readable predictions
        file_name = f'{self.add_step_name(run_id)}_{season}_{segment}-{self.config.segment_count}.csv'
        save_path = os.path.join(self.config.local_root, self.config.preds_path, run_id)

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        # save preds
        df_preds.to_csv(os.path.join(save_path, file_name),  index=False)
        logging.info(f'Saved predictions to {os.path.join(save_path, file_name)}')

        # do postprocessing and profit analysis
        if self.config.target_col.lower() == 'cover':
            if future:
                df_avg = mu.Util.unswap_results(df_preds, report_cols=list(df_preds.columns))
                df_avg.to_csv(f"{save_path}/avg_results_{self.add_step_name(run_id)}_"
                              f"{season}_{segment}-{self.config.segment_count}.csv", index=False)
                return self.aml, df_preds, df_avg
            else:
                df_avg, df_profit, profit_fig = mu.Util.graph_profit_risk(df_preds, report_cols=self.config.report_cols)
                df_profit.to_csv(f"{save_path}/profit_{self.add_step_name(run_id)}_"
                                 f"{season}_{segment}-{self.config.segment_count}.csv", index=False)
                profit_fig.savefig(f"{save_path}/graph_{self.add_step_name(run_id)}_"
                                   f"{season}_{segment}-{self.config.segment_count}.png", dpi=320)
                df_avg.to_csv(f"{save_path}/avg_results_{self.add_step_name(run_id)}_"
                              f"{season}_{segment}-{self.config.segment_count}.csv", index=False)
                return self.aml, df_preds, df_avg, df_profit, profit_fig
        else:
            return self.aml, df_preds

    # -----------------------------------------------------------------------------------
    # Combines the predictions from a run into a single df
    # -----------------------------------------------------------------------------------
    def combine_predictions(self):

        preds_list = [f for f in os.listdir(self.config.preds_path + self.config.run_id)]

        df = pd.DataFrame()

        for i, preds in enumerate(preds_list):
            df_tmp = pd.read_csv(f'{self.config.preds_path}{self.config.run_id}/{preds}')
            df = pd.concat([df_tmp, df])

        return df

    # -----------------------------------------------------------------------------------
    # create a kaggle submission file
    # -----------------------------------------------------------------------------------
    def create_kaggle_submission(self, df):

        df['Id'] = df.apply(
            lambda x: (str(int(x['Season'])) + '_' + str(int(x['T1_TeamID'])) + '_' + str(int(x['T2_TeamID']))), axis=1)
        df = df.rename(columns={'p1': 'Pred'})
        df = df[['Id', 'Pred']]

        df.to_csv(f"{self.config.submit_path}{self.config.run_id}.csv", index=False)

        return df

    # -----------------------------------------------------------------------------------
    # makes the message list for the record step
    # -----------------------------------------------------------------------------------
    def _get_report_msgs(self, aml):

        # list of messages to log
        msgs = []

        # all reports are on the best model
        best_model = aml.get_best_model()
        stats = best_model.model_performance(valid=True)

        msgs += ['-------------------------------------------------------------------------------------']
        msgs += [f"| Run ID: {self.config.run_id}"]
        msgs += [
            f"| Train start: {self.config.train_start}   Train end: {self.config.train_end}   Predicting: {self.config.test_start} >> {self.config.test_end}"]
        msgs += ["|"]
        msgs += [f"| {stats}"]
        msgs += ['-------------------------------------------------------------------------------------']

        return msgs

    def _annotate_analysis_info(self, df, future=False):

        local_path = os.path.join(self.config.local_root, self.config.data_path)

        df_annotation_source = pd.read_csv(os.path.join(local_path, 'MMLast_Pred.csv')) if future else pd.read_csv(
                                           os.path.join(local_path, 'MMLast_Test.csv'))
        df_annotation_source = df_annotation_source[self.config.report_cols]
        df = df_annotation_source.join(df)
        df = MMLookup(source='teams', path=self.config.data_path, local_root=self.config.local_root,
                      gcs_root=self.config.gcs_root).replace_ids(df, 'TeamID', 'TeamName')

        return df

    def update_model_db(self, model=None, model_path=None, season=None, segment=None, segment_count=None,
                        target_name=None, description=None):

        # set our key fields
        season = self.config.test_start[0] if season is None else season
        segment = self.config.test_start[1] if segment is None else segment
        target_name = self.config.target_col if target_name is None else target_name
        segment_count = self.config.segment_count if segment_count is None else segment_count
        description = self.config.description if description is None else description

        # get the model_path if it's passed in
        if model_path is not None:
            run_id = pathlib.Path(model_path).parts[-3]
            model_file = pathlib.Path(model_path).parts[-1]
            model_path = f"{self.config.model_path}{run_id}/Step_000/{model_file}"

        # get the type of score and pull the current best score from model_db
        if self.config.target_type == "continuous":
            metric = "MAE"
            model_score = model.mae(xval=True)

        elif self.config.target_type == "multiclass":
            metric = "logloss"
            model_score = model.logloss(xval=True)

        best_score = self.model_db.loc[(self.model_db['Season'] == season) &
                                       (self.model_db['Segment'] == segment) &
                                       (self.model_db['Target'] == target_name) &
                                       (self.model_db['Description'] == description) &
                                       (self.model_db['Segment_Count'] == segment_count) &
                                       (self.model_db['Metric'] == metric)]["Score"].reset_index(drop=True)


        # see if there is improvement
        if best_score.empty:
            best_score = model_score + 1
            drop = False
        else:
            best_score = best_score[0]
            drop = True

        # if there is drop the current row from model_db and add the new row with improved metrics
        if best_score > model_score:
            if drop:
                self.model_db = self.model_db.drop(self.model_db[(self.model_db['Season'] == season) &
                                                                 (self.model_db['Segment'] == segment) &
                                                                 (self.model_db['Target'] == target_name) &
                                                                 (self.model_db['Description'] == description) &
                                                                 (self.model_db['Segment_Count'] == segment_count) &
                                                                 (self.model_db['Metric'] == metric)].index)

            row = pd.DataFrame({'Season': season, 'Segment': segment, 'Segment_Count': segment_count,
                                'Target': target_name, 'Description': description, "Metric": metric,
                                'Score': model_score, 'Model_File': model_path}, index=[0])
            self.model_db = pd.concat([self.model_db, row], ignore_index=True)
            self.model_db.to_csv(f"{self.config.model_path}model_database.csv", index=False)
            logging.info(f'Season: {season} | Segment: {segment} | Model: {model_file} >> saved to model database..')

        else:
            logging.info(f'Season: {season} | Segment: {segment} | Model: {model_file} >> Is no improvement, '
                         f'not added to model_database.csv..')


    def open_model_file(self, run_id=None, model_file=None, gcs_root='ops/', local_root='./'):

        p = f"{self.config.model_path}{run_id}/Step_000/"
        file_name = os.path.join(p, model_file)
        local_name = os.path.join(local_root, file_name)
        gcs_name = os.path.join(gcs_root, file_name)

        mu.Util.make_paths([os.path.join(local_root, p)])

        h2o.no_progress()

        if not os.path.exists(file_name):
            if not mu.Util.download(gcs_name=gcs_name, local_name=local_name):
                logging.error(f'Could not download {gcs_name}')

        if model_file[-3:] == "zip":
            model = h2o.import_mojo(local_name)
        else:
            model = h2o.load_model(local_name)

        h2o.show_progress()

        return model

